# Author: Robert H Cudmore
# Date: 20190719

from curses.panel import bottom_panel
import os, sys
#import traceback # to print call stack
from functools import partial
from collections import OrderedDict
import platform
import pathlib
from datetime import datetime

from typing import Union, Dict, List, Tuple

import pandas as pd

import pyqtgraph as pg

# v1
#import qdarkstyle  # the one we originally used

# v2
import qdarktheme
# Enable HiDPI.
qdarktheme.enable_hi_dpi()

import webbrowser  # to open online help

from PyQt5 import QtCore, QtWidgets, QtGui

import sanpy._util
import sanpy.interface
import sanpy.interface.preferences

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)
# This causes mkdocs to infinite recurse when running locally as 'mkdocs serve'
#logger.info('SanPy app.py is starting up')

# turn off qdarkstyle logging
import logging
# logging.getLogger('qdarkstyle').setLevel(logging.WARNING)

# turn off numexpr 'INFO' logging
logging.getLogger('numexpr').setLevel(logging.WARNING)

#basedir = os.path.dirname(__file__)
#print('sanpy_app basedir:', basedir)

class SanPyWindow(QtWidgets.QMainWindow):

    #TODO: define parameter block for all signals.
    
    signalSetXAxis = QtCore.pyqtSignal(object)
    """Emit set axis."""

    signalSwitchFile = QtCore.pyqtSignal(object, object)
    """Emit on switch file."""

    signalSelectSweep = QtCore.pyqtSignal(object, object)  # (ba, sweepNumber)
    """Emit set sweep."""

    signalUpdateAnalysis = QtCore.pyqtSignal(object)
    """Emit on detect."""

    signalSelectSpike = QtCore.pyqtSignal(object)
    """Emit spike selection."""

    signalSelectSpikeList = QtCore.pyqtSignal(object)
    """Emit spike list selection."""

    def __init__(self, path=None, parent=None):
        """
        Parameters
        ----------
        path : str
            Full path to folder with raw files (abf,csv,tif).
        """

        super().__init__(parent)

        self.setAcceptDrops(True)

        # TODO: if first time run (no <user>/Documents/SanPy) then warn user to quit and restart

        # create directories in <user>/Documents and add to python path
        firstTimeRunning = sanpy._util.addUserPath()
        if firstTimeRunning:
            logger.info('  We created <user>/Documents/Sanpy and need to restart')

        self._fileLoaderDict = sanpy.fileloaders.getFileLoaders()
        self._detectionClass = sanpy.bDetection()

        # create an empty model for file list
        dfEmpty = pd.DataFrame(columns=sanpy.analysisDir.sanpyColumns.keys())
        self.myModel = sanpy.interface.bFileTable.pandasModel(dfEmpty)

        self.fileFromDatabase = True  # if False then from folder

        self.startSec = None
        self.stopSec = None

        #TODO: put font size in options file
        # removed mar 26 2023
        '''
        myFontSize = 10
        myFont = self.font();
        myFont.setPointSize(myFontSize)
        self.setFont(myFont)
        '''

        if path is not None and os.path.isdir(path):
            windowTitle = f'SanPy {path}'
        else:
            windowTitle = 'SanPy'
        self.setWindowTitle(windowTitle)

        self._rowHeight = 11

        # path to loaded folder (using bAnalysisDir)
        self.configDict : sanpy.interface.preferences = sanpy.interface.preferences(self)
        self.myAnalysisDir = None
        lastPath = self.configDict.getMostRecentFolder()
        logger.info(f'  preferences lastPath is "{lastPath}"')
        if path is not None:
            self.path = path
        elif lastPath is not None and os.path.isdir(lastPath):
            self.path = lastPath
        else:
            self.path = None

        #TODO: refactor dark to light theme
        self.useDarkStyle = self.configDict['useDarkStyle']
        #self.useDarkStyle = True
        self.toggleStyleSheet(self.useDarkStyle, buildingInterface=True)

        #
        # set window geometry
        self.setMinimumSize(640, 480)
        self.left = self.configDict['windowGeometry']['x']
        self.top = self.configDict['windowGeometry']['y']
        self.width = self.configDict['windowGeometry']['width']
        self.height = self.configDict['windowGeometry']['height']
        self.setGeometry(self.left, self.top, self.width, self.height)

        self.myPlugins = sanpy.interface.bPlugins(sanpyApp=self)

        # order matter, _buildMenus uses objects created in _buildUI
        self._buildUI()
        self._buildMenus()

        # 20210803, loadFolder was above? Still works down here
        # needed to update detection widget after buildUI()
        if self.path is not None and len(self.path)>0:
            self.slot_loadFolder(self.path)

        # self.raise_()  # bring to front, raise is a python keyword
        # self.activateWindow()  # bring to front

        self.slot_updateStatus('Ready')
        logger.info('SanPy started')

    def getDetectionClass(self):
        return self._detectionClass

    def getFileLoaderDict(self):
        return self._fileLoaderDict
    
    def dragEnterEvent(self, event):
        logger.info('')
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        logger.info('')
        folders = [u.toLocalFile() for u in event.mimeData().urls()]
        if len(folders) > 1:
            return
        oneFolder = folders[0]
        if os.path.isdir(oneFolder):
            self.slot_loadFolder(path=oneFolder)

    def _promptIfDirty(self) -> bool:
        """Prompt user if there is unsaved analysis.
        
        If this return False, do not proceed with caller action.
        e.g. on 'load folder'

        If this returns True then proceed with caller action
        """
        acceptAndContinue = True
        if self.myAnalysisDir is not None:
            tableIsDirty = self.myAnalysisDir.isDirty
            analysisIsDirty = self.myAnalysisDir.hasDirty()
            if tableIsDirty or analysisIsDirty:
                userResp = sanpy.interface.bDialog.yesNoCancelDialog(
                                    'There is analysis that is not saved.\nDo you want to save?')
                if userResp == QtWidgets.QMessageBox.Yes:
                    self.saveFilesTable()
                    acceptAndContinue = True
                elif userResp == QtWidgets.QMessageBox.No:
                    acceptAndContinue = True
                else:  # userResp == QtWidgets.QMessageBox.Cancel:
                    acceptAndContinue = False
        return acceptAndContinue

    def closeEvent(self, event):
        """Called when user closes main window or selects quit.
        
        Parameters
        ----------
        event : PyQt5.QtGui.QCloseEvent
        """

        logger.info(event)

        # check if our table view has been edited by user and warn
        doQuit = True
        alreadyAsked = False
        # self.myAnalysisDir is only defined after we load a folder
        if self.myAnalysisDir is not None:
            tableIsDirty = self.myAnalysisDir.isDirty
            analysisIsDirty = self.myAnalysisDir.hasDirty()
            if tableIsDirty or analysisIsDirty:
                alreadyAsked = True
                userResp = sanpy.interface.bDialog.yesNoCancelDialog('There is analysis that is not saved.\nDo you want to save?')
                if userResp == QtWidgets.QMessageBox.Yes:
                    self.saveFilesTable()
                    event.accept()
                elif userResp == QtWidgets.QMessageBox.No:
                    event.accept()
                else:
                    event.ignore()
                    doQuit = False
        if doQuit:
            if not alreadyAsked:
                userResp = sanpy.interface.bDialog.okCancelDialog('Are you sure you want to quit SanPy?', informativeText=None)
                if userResp == QtWidgets.QMessageBox.Cancel:
                    event.ignore()
                    doQuit = False

            if doQuit:
                logger.info('SanPy is quiting')
                QtCore.QCoreApplication.quit()

    def getOptions(self):
        return self.configDict

    def getWindowGeometry(self):
        """Get the current window position.
        """
        myRect = self.geometry()
        left = myRect.left()
        top = myRect.top()
        width = myRect.width()
        height = myRect.height()
        return left, top, width, height

    def toggleStyleSheet(self, doDark=None, buildingInterface=False):
        logger.info('')
        if doDark is None:
            doDark = not self.useDarkStyle
        self.useDarkStyle = doDark
        if doDark:
            # v1
            #self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt5'))
            # v2
            qdarktheme.setup_theme('dark')

            pg.setConfigOption('background', 'k')
            pg.setConfigOption('foreground', 'w')
        else:
            # v1
            # self.setStyleSheet('')
            # v2
            qdarktheme.setup_theme('light')

            pg.setConfigOption('background', 'w')
            pg.setConfigOption('foreground', 'k')

        self.configDict['useDarkStyle'] = self.useDarkStyle

        if not buildingInterface:
            #self.myScatterPlotWidget.defaultPlotLayout()
            #self.myScatterPlotWidget.buildUI(doRebuild=True)
            self.myDetectionWidget.mySetTheme()

        if buildingInterface:
            pass
        else:
            pass
            # msg = QtWidgets.QMessageBox()
            # msg.setIcon(QtWidgets.QMessageBox.Warning)
            # msg.setText("Theme Changed")
            # msg.setInformativeText('Please restart SanPy for changes to take effect.')
            # msg.setWindowTitle("Theme Changed")
            # retval = msg.exec_()

            #self.configDict.save()

    def slot_loadFolder(self, path='', folderDepth=None):
        """Load a folder of raw data files.

        Parameters
        ----------
        path : str
        folderDepth : int or None
        """

        if folderDepth is None:
            # get the depth from file list widget
            folderDepth = self._fileListWidget.getDepth()

        logger.info(f'Loading depth:{folderDepth} path: {path}')

        # if folder is already loaded, ask to save
        acceptAndContinue = self._promptIfDirty()
        if not acceptAndContinue:
            return

        # ask user for folder
        if path is None or isinstance(path,bool) or len(path)==0:
            path = str(QtWidgets.QFileDialog.getExistingDirectory(self, "Select Directory With Recordings"))
            if len(path) == 0:
                return
        elif os.path.isdir(path):
            pass
        else:
            logger.warning(f'    Did not load path "{path}"')
            return

        self.path = path # path to loaded bAnalysisDir folder

        #logger.info(f'Loading path: {path}')

        # will create/load hd5 file for folder
        self.myAnalysisDir = sanpy.analysisDir(path, folderDepth=folderDepth, myApp=self)

        # set myAnalysisDir to file list model
        self.myModel = sanpy.interface.bFileTable.pandasModel(self.myAnalysisDir)
        #self.myModel.signalMyDataChanged.connect(self.slot_dataChanged)
        #self.myModel.signalMyDataChanged.connect(self.myDetectionWidget.slot_dataChanged)

        #try:
        if 1:
            self._fileListWidget.mySetModel(self.myModel)
            self.myModel.signalMyDataChanged.connect(self.myDetectionWidget.slot_dataChanged)
        '''
        except (AttributeError) as e:
            # needed when we call loadFolder from __init__
            # logger.warning('OK: no tableView during load folder')
            pass
        '''

        # set window title
        if self.path is not None and os.path.isdir(self.path):
            windowTitle = f'SanPy: {self.path}'
        else:
            windowTitle = 'SanPy'
        self.setWindowTitle(windowTitle)

        # add to preferences recent folders
        self.configDict.addFolder(path)

        # save preferences
        #self.configDict.save()
        
    '''
    def slot_dataChanged(self, columnName, value, rowDict):
        """User has edited main file table.
        Update detection widget for columns (Start(s), Stop(s), dvdtThreshold, mvThreshold)
        """
        logger.info(f'{columnName} {value}')
        print('  ', rowDict)
    '''
    def selectSpike(self, spikeNumber, doZoom=False):
        eDict = {
                'spikeNumber': spikeNumber,
                'doZoom': doZoom,
                'ba': self.get_bAnalysis(),
                }
        self.signalSelectSpike.emit(eDict)

    def selectSpikeList(self, spikeList : List[int], doZoom : bool = False):
        eDict = {
                'spikeList': spikeList,
                'doZoom': doZoom,
                'ba': self.get_bAnalysis(),
                }
        self.signalSelectSpikeList.emit(eDict)

    def mySignal(self, this, data=None):
        """Receive signals from children widgets.

        Parameters
        ----------
        this : str
            The signal name
        data : type depends on signal (this)
            For example, signal 'set x axis' uses data=[min,max]
        """
        #print('=== sanpy_app.mySignal() "' + this +'"')

        if this == 'select spike':
            logger.warning('\n\nTODO: GET RID OF "select spike"\n\n')
            spikeNumber = data['spikeNumber']
            doZoom = data['isShift']
            self.selectSpike(spikeNumber, doZoom=doZoom)
            #self.signalSelectSpike.emit(data)

        elif this == 'set x axis':
            logger.info(f'"set x axis" {data}')

            self.startSec = data[0]
            self.stopSec = data[1]
            # old
            #self.myScatterPlotWidget.selectXRange(data[0], data[1])
            # new
            self.signalSetXAxis.emit([data[0], data[1]])  # emits to scatter plot ONLY

        elif this == 'set full x axis':
            self.startSec = 0
            if self.get_bAnalysis() is not None:
                self.stopSec = self.get_bAnalysis().fileLoader.recordingDur
            else:
                self.stopSec = None
            logger.info(f'"set full x axis" {self.startSec} {self.stopSec}')
            self.signalSetXAxis.emit([self.startSec, self.stopSec])  # emits to scatter plot ONLY

        elif this == 'cancel all selections':

            self.selectSpike(None)
            self.selectSpikeList([])

        else:
            logger.warning(f'Did not understand this: "{this}"')

    def old_keyPressEvent(self, event):
        """Respond to key press
        
        TODO: does not seem to work?
        """
        key = event.key()
        text = event.text()
        logger.info(f'text:{text} key:{key} event:{event}')

        # handled in bDetectionWidget
        # set full axis
        #if key in [70, 82]: # 'r' or 'f'
        # if key in [QtCore.Qt.Key.Key_R, QtCore.Qt.Key.Key_F]:
        #     self.myDetectionWidget.setAxisFull()

        '''
        if key in [QtCore.Qt.Key.Key_P]: # 'r' or 'f'
            self.myDetectionWidget.myPrint()
        '''

        # handled in bDetectionWidget
        # cancel all selections
        # if key == QtCore.Qt.Key.Key_Escape:
        #     self.mySignal('cancel all selections')

        # handled in bDetectionWidget
        # hide detection widget
        # if text == 'h':
        #     if self.myDetectionWidget.detectToolbarWidget.isVisible():
        #         self.myDetectionWidget.detectToolbarWidget.hide()
        #     else:
        #         self.myDetectionWidget.detectToolbarWidget.show()

        # user can copy this to the clipboard
        # print file list model
        # this is df updated as user updates table
        # if text == 'p':
        #     print(self.myModel)
        #     print(self.myModel._data)

        #
        event.accept()

    def get_bAnalysis(self):
        return self.myDetectionWidget.ba

    def getSelectedFileDict(self):
        """
        Used by detection widget to get info in selected file.

        todo: remove, pass this dict in signal emit from file table
        """
        
        fileDict = self._fileListWidget.getTableView().getSelectedRowDict()
        return fileDict

        # was this
        '''
        selectedRows = self._fileListWidget.selectionModel().selectedRows()
        if len(selectedRows) == 0:
            return None
        else:
            selectedItem = selectedRows[0]
            selectedRow = selectedItem.row()

        rowDict = self.myModel.myGetRowDict(selectedRow)
        return rowDict
        '''

    def slot_fileTableClicked(self, row, rowDict, selectingAgain):
        """Respond to selections in file table."""

        '''
        row (int):
        rowDict (dict):
        selectingAgain (bool): True if row was already selected
        '''

        if selectingAgain:
            self.slot_updateStatus(f'Refreshing file "{rowDict["File"]}"')
        else:
            self.slot_updateStatus(f'Loading file "{rowDict["File"]}" ... please wait')

        # TODO: try and remove this
        self.startSec = rowDict['Start(s)']
        self.stopSec = rowDict['Stop(s)']

        # This will load if necc, otherwise just fetch a pointer
        if self.myAnalysisDir is not None:
            ba = self.myAnalysisDir.getAnalysis(row) # if None then problem loading

            if ba is not None:
                self.signalSwitchFile.emit(ba, rowDict)
                if selectingAgain:
                    pass
                else:
                    self.slot_updateStatus(f'Loaded file "{ba.fileLoader.filename}"')# this will load ba if necc

    def _buildMenus(self):

        mainMenu = self.menuBar()

        # load
        loadFolderAction = QtWidgets.QAction('Load Folder ...', self)
        loadFolderAction.setShortcut('Ctrl+O')
        loadFolderAction.triggered.connect(self.slot_loadFolder)

        # open recent (submenu)
        self.openRecentMenu = QtWidgets.QMenu('Load Recent ...')
        self.openRecentMenu.aboutToShow.connect(self._refreshOpenRecent)

        saveDatabaseAction = QtWidgets.QAction('Save Folder Analysis', self)
        saveDatabaseAction.setShortcut('Ctrl+S')
        saveDatabaseAction.triggered.connect(self.saveFilesTable)

        #buildDatabaseAction = QtWidgets.QAction('Build Big Database ...', self)
        #buildDatabaseAction.triggered.connect(self.buildDatabase)

        savePreferencesAction = QtWidgets.QAction('Save Preferences', self)
        savePreferencesAction.triggered.connect(self.configDict.save)

        showLogAction = QtWidgets.QAction('Show Log', self)
        showLogAction.triggered.connect(self.openLog)

        fileMenu = mainMenu.addMenu('&File')
        fileMenu.addAction(loadFolderAction)
        fileMenu.addMenu(self.openRecentMenu)
        fileMenu.addSeparator()
        fileMenu.addAction(saveDatabaseAction)
        fileMenu.addSeparator()
        #fileMenu.addAction(buildDatabaseAction)
        #fileMenu.addSeparator()
        fileMenu.addAction(savePreferencesAction)
        fileMenu.addSeparator()
        fileMenu.addAction(showLogAction)

        # view menu to toggle widgets on/off
        self.viewMenu = mainMenu.addMenu('&View')
        self.viewMenu.aboutToShow.connect(self._refreshViewMenu)
        self._refreshViewMenu()
        #self._populateViewMenu()

        # self.windowsMenu = mainMenu.addMenu('&Window')
        # self.windowsMenu.aboutToShow.connect(self._refreshWindowsMenu)
        # self._refreshWindowsMenu()

        #
        # plugins menu
        pluginsMenu = mainMenu.addMenu('&Plugins')
        # list of plugin names
        # pluginList = self.myPlugins.pluginList()
        # each key is the name of theplugin
        pluginDict = self.myPlugins.pluginDict
        #print('pluginDict:', pluginDict)
        _foundUserPlugin = False
        for __humanName,v in pluginDict.items():
            if not v['showInMenu']:
                continue
            
            #logger.info(f'adding plugin: {plugin}')
            
            sanpyPluginAction = QtWidgets.QAction(__humanName, self)

            # TODO: Add spacer between system and user plugins
            #fileMenu.addSeparator()

            '''
            type = self.myPlugins.getType(plugin)
            if type == 'system':
                print(plugin, 'system -->> bold')
                f = sanpyPluginAction.font()
                f.setBold(True);
                f.setItalic(True);
                sanpyPluginAction.setFont(f);
            '''

            #print(v['type'])
            if v['type'] == 'user':
                if not _foundUserPlugin:
                    pluginsMenu.addSeparator()
                    _foundUserPlugin = True
                _font = sanpyPluginAction.font()
                #_font.setBold(True)    
                _font.setItalic(True)    
                sanpyPluginAction.setFont(_font)

            sanpyPluginAction.triggered.connect(lambda checked, pluginName=__humanName: self.sanpyPlugin_action(pluginName))
            pluginsMenu.addAction(sanpyPluginAction)

        '''
        pluginDir = os.path.join(self._getBundledDir(), 'plugins', '*.txt')
        pluginList = glob.glob(pluginDir)
        logger.info(f'pluginList: {pluginList}')
        pluginsMenu = mainMenu.addMenu('&Plugins')
        oneAction = 'plotRecording'
        sanpyPluginAction = QtWidgets.QAction(oneAction, self)
        #sanpyPluginAction.triggered.connect(self.sanpyPlugin_action)
        sanpyPluginAction.triggered.connect(lambda checked, oneAction=oneAction: self.sanpyPlugin_action(oneAction))
        pluginsMenu.addAction(sanpyPluginAction)
        '''

        #
        # a dynamic menu to show open plugins
        # self.windowsMenu = mainMenu.addMenu('&Windows')
        # self.windowsMenu.aboutToShow.connect(self._populateOpenPlugins)

        '''
        # windows menu to toggle scatter plot widget
        windowsMenu = mainMenu.addMenu('&Windows')
        mainWindowAction = QtWidgets.QAction('Main', self)
        #
        openScatterAction = QtWidgets.QAction('Scatter Plot', self)
        openScatterAction.triggered.connect(self.openScatterWindow)
        #mainWindowAction.triggered.connect(self.toggleStyleSheet)
        mainWindowAction.setCheckable(True)
        mainWindowAction.setChecked(True)
        windowsMenu.addAction(mainWindowAction)
        windowsMenu.addAction(openScatterAction)
        '''

        # help menu
        helpMenu = mainMenu.addMenu('&Help')
        name = 'SanPy Help (Opens In Browser)'
        action = QtWidgets.QAction(name, self)
        action.triggered.connect(partial(self._onHelpMenuAction, name))
        helpMenu.addAction(action)

    def _onHelpMenuAction(self, name : str):
        if name == 'SanPy Help (Opens In Browser)':
            url = 'https://cudmore.github.io/SanPy/desktop-application'
            webbrowser.open(url, new=2)
        
    def _refreshOpenRecent(self):
        """Dynamically generate the open recent file menu.
        """
        self.openRecentMenu.clear()
        for recentFolder in self.configDict.getRecentFolder():

            loadFolderAction = QtWidgets.QAction(recentFolder, self)
            #loadFolderAction.setShortcut('Ctrl+O')
            loadFolderAction.triggered.connect(partial(self.slot_loadFolder, recentFolder))

            self.openRecentMenu.addAction(loadFolderAction)

    def _refreshWindowsMenu(self):
        self.windowsMenu.clear()

        name = 'SanPy Window'
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(self.isVisible())
        action.triggered.connect(partial(self._windowsMenuAction, 'xxx', name))
        self.windowsMenu.addAction(action)

    def _refreshViewMenu(self):
        """Dynamically create the main 'View' menu each time it is selected.
        """
        
        self.viewMenu.clear()

        self.viewMenu.addSeparator()

        key1 = 'filePanels'
        name = 'File Panel'
        checkedMainPanel = self.configDict[key1]['File Panel']
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checkedMainPanel)
        action.setShortcut(QtGui.QKeySequence('F'))
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        self.viewMenu.addSeparator()

        key1 = 'detectionPanels'

        name = 'Detection Panel'
        checkedMainPanel = self.configDict[key1]['Detection Panel']
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checkedMainPanel)
        action.setShortcut(QtGui.QKeySequence('D'))
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        self.viewMenu.addSeparator()

        name = 'Detection'
        checked = self.configDict['detectionPanels'][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        action.setEnabled(checkedMainPanel)
        self.viewMenu.addAction(action)

        name = 'Display'
        checked = self.configDict['detectionPanels'][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        action.setEnabled(checkedMainPanel)
        self.viewMenu.addAction(action)

        name = 'Plot Options'
        checked = self.configDict['detectionPanels'][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        action.setEnabled(checkedMainPanel)
        self.viewMenu.addAction(action)

        # mar 11
        name = 'Set Spikes'
        checked = self.configDict['detectionPanels'][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        action.setEnabled(checkedMainPanel)
        self.viewMenu.addAction(action)

        self.viewMenu.addSeparator()

        # traces (globalvm, dvdt, dac)
        key1 = 'rawDataPanels'

        name = 'Full Recording'
        checked = self.configDict[key1][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        name = 'Derivative'
        checked = self.configDict[key1][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        name = 'DAC'
        checked = self.configDict[key1][name]
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        self.viewMenu.addSeparator()

        # show/hide plugin 1/2 dock widgets
        key1 = 'pluginDocks'
        name = 'Plugins 1'
        #checkedMainPanel = self.configDict[key1]['File Panel']
        _isVisible = self.pluginDock1.isVisible()  # assuming we _buildUI
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(_isVisible)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)

        '''
        key1 = 'pluginDocks'
        name = 'Plugins 2'
        _isVisible = self.pluginDock2.isVisible()  # assuming we _buildUI
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(_isVisible)
        action.triggered.connect(partial(self._viewMenuAction, key1, name))
        self.viewMenu.addAction(action)
        '''

        self.viewMenu.addSeparator()

        #TODO: refactor switching themes
        name = 'Dark Theme'
        checked = self.useDarkStyle
        action = QtWidgets.QAction(name, self, checkable=True)
        action.setChecked(checked)
        #action.setEnabled(False)
        action.triggered.connect(partial(self._viewMenuAction, "Dark Theme", name))
        self.viewMenu.addAction(action)



    def _windowsMenuAction(self, key1, name, isChecked):
        self.setVisible(not self.isVisible())

    def _viewMenuAction(self, key1, name, isChecked):
        """Respond to user selection in view menu.
        """
        logger.info(f'{key1}, {name}, {isChecked}')
        
        try:
            self.configDict[key1][name] = isChecked
        except (KeyError) as e:
            pass

        if key1 == 'filePanels':
            self.toggleInterface(name, isChecked)

        elif key1 == 'pluginDocks':
            self.toggleInterface(name, isChecked)

        elif key1 == 'rawDataPanels':
            #self.toggleInterface(name, isChecked)
            self.myDetectionWidget.toggleInterface(name, isChecked)

        elif key1 == 'detectionPanels':
            self.myDetectionWidget.toggleInterface(name, isChecked)

        elif key1 == 'Dark Theme':
            # doDark = not self.useDarkStyle
            # self.toggleStyleSheet(doDark=doDark)
            self.toggleStyleSheet()
        else:
            logger.warning(f'  no action for key1: "{key1}"')

    def toggleInterface(self, name, on):
        """Toggle named interface widgets show and hide.
        
        Parameters
        ----------
        name : str
            Name of the widget
        on : bool
            If True then show, otherwise hide.
        """
        if name == 'File Panel':
            self._fileListWidget.setVisible(on)
            # if on:
            #     self.fileDock.show()
            # else:
            #     self.fileDock.hide()
        elif name == 'Plugins 1':
            if on:
                self.pluginDock1.show()
            else:
                self.pluginDock1.hide()
        elif name == 'Plugins 2':
            if on:
                self.pluginDock2.show()
            else:
                self.pluginDock2.hide()

    def old_populateOpenPlugins(self):
        """Depreciated. Was to be a dynamic menu to show open plugins.
        """
        self.windowsMenu.clear()
        actions = []
        for plugin in self.myPlugins._openSet:
            name = plugin.myHumanName
            windowTitle = plugin.windowTitle
            action = QtWidgets.QAction(windowTitle, self)
            action.triggered.connect(partial(self.old_showOpenPlugin, name, plugin, windowTitle))
            actions.append(action)
        self.windowsMenu.addActions(actions)

    def old_showOpenPlugin(self, name, plugin, windowTitle, selected):
        logger.info(name)
        logger.info(plugin)
        logger.info(windowTitle)
        logger.info(selected)
        plugin.bringToFront()

    def _buildUI(self):
        """"
        File List : sanpy.interface.fileListWidget
        Detection Widget: sanpy.interface.bDetectionWidget
        """
        #self.toggleStyleSheet(buildingInterface=True)

        self.statusBar = QtWidgets.QStatusBar()
        self.setStatusBar(self.statusBar)

        # typical wrapper for PyQt, we can't use setLayout(), we need to use setCentralWidget()
        _mainWidget = QtWidgets.QWidget()
        _mainVLayout = QtWidgets.QVBoxLayout()
        _mainWidget.setLayout(_mainVLayout)
        self.setCentralWidget(_mainWidget)

        #
        # list of files (in a dock)
        self._fileListWidget = sanpy.interface.fileListWidget(self.myModel)
        #self._fileListWidget.signalUpdateStatus.connect(self.slot_updateStatus)  # never used
        self._fileListWidget.signalLoadFolder.connect(self.slot_loadFolder)
        self._fileListWidget.getTableView().signalSelectRow.connect(self.slot_fileTableClicked)
        #self._fileListWidget.getTableView().signalSetDefaultDetection.connect(self.slot_setDetectionParams)

        # update dvdtThreshold, mvThreshold Start(s), Stop(s)
        self.signalUpdateAnalysis.connect(self._fileListWidget.getTableView().slot_detect)
        #self.myDetectionWidget.signalDetect.connect(self._fileListWidget.slot_detect)

        # file list as a widget
        _mainVLayout.addWidget(self._fileListWidget)

        # file list as a dock
        # self.fileDock = QtWidgets.QDockWidget('Files',self)
        # self.fileDock.setWidget(self._fileListWidget)
        # self.fileDock.setFloating(False)
        # self.fileDock.visibilityChanged.connect(self.on_fileDock_visibilityChanged)
        # self.fileDock.topLevelChanged.connect(self.slot_topLevelChanged)
        # self.fileDock.setAllowedAreas(QtCore.Qt.TopDockWidgetArea | QtCore.Qt.BottomDockWidgetArea)
        # self.fileDock.dockLocationChanged.connect(partial(self.slot_dockLocationChanged, self.fileDock))
        # self.addDockWidget(QtCore.Qt.TopDockWidgetArea, self.fileDock)

        #
        # Detection widget
        baNone = None
        self.myDetectionWidget = sanpy.interface.bDetectionWidget(ba=baNone, mainWindow=self)
        
        # show/hide
        on = self.configDict['detectionPanels']['Detection Panel']
        self.myDetectionWidget.toggleInterface('Detection Panel', on)

        _mainVLayout.addWidget(self.myDetectionWidget)

        # myDetectionWidget listens to self
        self.signalSwitchFile.connect(self.myDetectionWidget.slot_switchFile)
        self.signalSelectSpike.connect(self.myDetectionWidget.slot_selectSpike)
        self.signalSelectSpikeList.connect(self.myDetectionWidget.slot_selectSpikeList)
        self.signalUpdateAnalysis.connect(self.myDetectionWidget.slot_updateAnalysis)

        # self listens to myDetectionWidget
        self.myDetectionWidget.signalSelectSpike.connect(self.slot_selectSpike)
        self.myDetectionWidget.signalSelectSpikeList.connect(self.slot_selectSpikeList)
        self.myDetectionWidget.signalSelectSweep.connect(self.slot_selectSweep)
        self.myDetectionWidget.signalDetect.connect(self.slot_detect)

        # detection widget is persistent
        # removed mar 26 2023
        # self.setCentralWidget(self.myDetectionWidget)
        
        # TODO: detection widget as a dock
        #  detection widget has left panel of controls and right panel of plots
        #  just make left controls a dock widget
        # self.detectionDock = QtWidgets.QDockWidget('Detection',self)
        # self.detectionDock.setWidget(self.myDetectionWidget)
        # self.detectionDock.setFloating(False)
        # self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.detectionDock)

        #self.setLayout(_mainVLayout)

        #
        # 2x docks for plugins
        self.myPluginTab1 = QtWidgets.QTabWidget()
        self.myPluginTab1.setMovable(True)
        self.myPluginTab1.setTabsClosable(True)
        self.myPluginTab1.tabCloseRequested.connect(partial(self.slot_closeTab, sender=self.myPluginTab1))
        self.myPluginTab1.currentChanged.connect(partial(self.slot_changeTab, sender=self.myPluginTab1))
        # re-wire right-click
        self.myPluginTab1.setContextMenuPolicy(QtCore.Qt.CustomContextMenu);
        self.myPluginTab1.customContextMenuRequested.connect(partial(self.on_plugin_contextMenu, sender=self.myPluginTab1))

        _openPlugins = self.configDict['pluginPanels'] # dict of
        for _openPlugin, _externalWindow in _openPlugins.items():
            _externalWindow = _externalWindow['externalWindow']
            logger.info(f'  _openPlugin:{_openPlugin} _externalWindow:{_externalWindow}')
            _oneOpenPlugin = self.myPlugins.runPlugin(_openPlugin, ba=None, show=False)
            if _oneOpenPlugin is not None:
                if _externalWindow:
                    self.sanpyPlugin_action(_openPlugin)
                else:
                    # on add tab, the QTabWIdget makes a copy !!!
                    self.myPluginTab1.addTab(_oneOpenPlugin, _oneOpenPlugin.myHumanName)

        self.pluginDock1 = QtWidgets.QDockWidget('Plugins 1',self)
        self.pluginDock1.setWidget(self.myPluginTab1)
        self.pluginDock1.setVisible(self.myPluginTab1.count()>0)
        self.pluginDock1.setFloating(False)
        self.pluginDock1.dockLocationChanged.connect(partial(self.slot_dockLocationChanged, self.pluginDock1))
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea,self.pluginDock1)

        #
        # tabs
        '''
        self.myPluginTab2 = QtWidgets.QTabWidget()
        self.myPluginTab2.setMovable(True)
        self.myPluginTab2.setTabsClosable(True)
        self.myPluginTab2.tabCloseRequested.connect(partial(self.slot_closeTab, sender=self.myPluginTab2))
        self.myPluginTab2.currentChanged.connect(partial(self.slot_changeTab, sender=self.myPluginTab2))
        # re-wire right-click
        self.myPluginTab2.setContextMenuPolicy(QtCore.Qt.CustomContextMenu);
        self.myPluginTab2.customContextMenuRequested.connect(partial(self.on_plugin_contextMenu, sender=self.myPluginTab2))

        self.pluginDock2 = QtWidgets.QDockWidget('Plugins 2',self)
        self.pluginDock2.setWidget(self.myPluginTab2)
        self.pluginDock2.setFloating(False)
        self.pluginDock2.hide() # initially hide 'Plugins 2'
        self.pluginDock2.dockLocationChanged.connect(partial(self.slot_dockLocationChanged, self.pluginDock2))
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.pluginDock2)
        '''

    def sanpyPlugin_action(self, pluginName):
        """Responds to main menu 'Plugins'.
        
        Run a plugin using curent ba.
        
        Notes:
            See also on_plugin_contextMenu for running a plugin in a tab
        """
        logger.info(f'opening pluginName: "{pluginName}"')
        ba = self.get_bAnalysis()
        
        # run the plugin
        _runningPlugin = self.myPlugins.runPlugin(pluginName, ba)

        ltwhTuple = _runningPlugin.getWindowGeometry()

        # add plugin to preferences so it opens next time we run the app
        if _runningPlugin is not None:
            self.configDict.addPlugin(_runningPlugin.getHumanName(),
                                      externalWindow=True,
                                      ltwhTuple = ltwhTuple)

    def slot_selectSpike(self, sDict):
        spikeNumber = sDict['spikeNumber']
        doZoom = sDict['doZoom']
        self.selectSpike(spikeNumber, doZoom)

    def slot_selectSpikeList(self, sDict):
        spikeList = sDict['spikeList']
        doZoom = sDict['doZoom']
        self.selectSpikeList(spikeList, doZoom)

    def slot_selectSweep(self, ba, sweepNumber):
        """Set view to new sweep.
        
        Parameters
        ----------
        ba : sanpy.bAnalysis
        sweepNumber : int
        """
        self.signalSelectSweep.emit(ba, sweepNumber)

    def saveFilesTable(self):
        """Save the folder hdf5 file.
        """
        #logger.info('')
        self.myAnalysisDir.saveHdf()
        self.slot_updateStatus(f'Save analysis for folder: {self.myAnalysisDir.path}')

    def slot_updateStatus(self, text : str):
        """Update the bottom status bar with new str
        """
        logger.info(text)
        
        # having trouble with an immediate update?
        self.statusBar.showMessage(text)
        self.statusBar.repaint()
        #self.statusBar.update()
        self.repaint()
        #self.update()
        QtWidgets.qApp.processEvents()

    def _old_slot_setDetectionParams(self, row, cellType):
        """Set detection parameters to presets.

        Parameters
        ----------
        row : int
            Selected row in file table
        cellType : str
            One of ('SA Node Params', 'Ventricular Params', 'Neuron Params')
        """
        logger.info(f'row:{row} cellType:{cellType}')
        self.myModel.mySetDetectionParams(row, cellType)

    def slot_detect(self, ba):
        """Respond to spike detection.
        
        Usually comes from the sanpy.interface.bDetectionWidget
        """

        # sweep number does not change
        self.signalUpdateAnalysis.emit(ba)

        self.slot_updateStatus(f'Detected {ba.numSpikes} spikes')

    def on_plugin_contextMenu(self, point, sender):
        """On right-click in dock, build a menu of plugins.

        On user selection, run the plugin in a tab.

        Notes:
            See also sanpyPlugin_action for running a plugin outside a tab (via main plugin menu)

        Parameters
        ----------
        point :QtCore.QPoint)
            Not used
        sender : QTabWidget
        """
        #logger.info(f'point:{point}, sender:{sender}')

        # list of available plugins
        pluginList = self.myPlugins.pluginList()

        contextMenu = QtWidgets.QMenu(self)

        for plugin in pluginList:
            contextMenu.addAction(plugin)

        # get current mouse/cursor position
        # not sure what 'point' parameter is?
        pos = QtGui.QCursor.pos()
        action = contextMenu.exec_(pos)

        if action is None:
            # no menu selected
            return

        pluginName = action.text()
        ba = self.get_bAnalysis()
        newPlugin = self.myPlugins.runPlugin(pluginName, ba, show=False)

        # only add if plugin wants to be shown
        if not newPlugin.getInitError() and newPlugin.getShowSelf():
            # add tab

            # 1) either this
            #newPlugin.insertIntoScrollArea()
            '''
            scrollArea = newPlugin.insertIntoScrollArea()
            if scrollArea is not None:
                newTabIndex = sender.addTab(scrollArea, pluginName)
            else:
                newTabIndex = sender.addTab(newPlugin, pluginName)
            '''
            # 2) or this
            #newTabIndex = sender.addTab(newPlugin, pluginName)  # addTab takes ownership
            newTabIndex = sender.addTab(newPlugin.getWidget(), pluginName)  # addTab takes ownership

            #widgetPointer = sender.widget(newTabIndex)
            #widgetPointer.insertIntoScrollArea()

            # bring tab to front
            sender.setCurrentIndex(newTabIndex)

            ltwhTuple = newPlugin.getWindowGeometry()

            if newPlugin is not None:
                self.configDict.addPlugin(newPlugin.getHumanName(),
                                        externalWindow=False,
                                        ltwhTuple=ltwhTuple)
                
    def slot_dockLocationChanged(self, dock, area):
        """Top level dock changed
        
        Parameters
        ----------
        dock : xxx
        area : enum QtCore.Qt.DockWidgetArea
            Basically left/top/right/bottom.

        Not triggered when user 'floats' a dock (See self.slot_topLevelChanged())
        """
        logger.info(f'not implemented, dock:"{dock.windowTitle()}" area enum: {area}')
        return

    def on_fileDock_visibilityChanged(self, visible : bool):
        """The file dock visibility was changed.
        """
        self._viewMenuAction('filePanels', 'File Panel', visible)

    def slot_topLevelChanged(self, topLevel):
        """
        topLevel (bool): True if the dock widget is now floating; otherwise False.

        This is triggered twice, once while dragging and once when finished
        """
        sender = self.sender()  # PyQt5.QtWidgets.QDockWidget
        logger.info(f'topLevel:{topLevel} sender:{sender}')
        return

    def slot_closeTab(self, index, sender):
        """Close an open plugin tab.

        Parameters
        ----------
        index : int
            The index into sender that gives us the tab, sender.widget(index)
        sender : PyQt5.QtWidgets.QTabWidget
            The tab group where a single tab was was closed
        """

        logger.info(f'index:{index} sender:{type(sender)}')

        # remove plugin from self.xxx
        # pluginInstancePointer is full class to actual plugin, like
        # sanpy.interface.plugins.detectionParams.detectionParams
        pluginInstancePointer = sender.widget(index)
        logger.info(f'  closing pluginInstancePointer:{type(pluginInstancePointer)}')

        # remove from preferences
        # if pluginInstancePointer is not None:
        #     self.configDict.removePlugin(pluginInstancePointer.getHumanName())

        self.myPlugins.slot_closeWindow(pluginInstancePointer)

        # remove the tab
        sender.removeTab(index)

    def slot_changeTab(self, index, sender):
        """User brought a different tab to the front

        Make sure only front tab (plugins) receive signals
        """
        logger.info(f'not implemented, index:{index} sender:{sender}')
        pass

    def openLog(self):
        """Open sanpy.log
        """
        logFilePath = sanpy.sanpyLogger.getLoggerFile()
        logFilePath = 'file://' + logFilePath
        url = QtCore.QUrl(logFilePath)
        QtGui.QDesktopServices.openUrl(url)

def runFft(sanpyWindow):
    logger.info('')
    sanpyWindow._fileListWidget._onLeftClick(0)
    sanpyWindow.myDetectionWidget.setAxis(2.1,156.8)

    ba = sanpyWindow.get_bAnalysis()
    pluginName = 'FFT'
    fftPlugin = sanpyWindow.myPlugins.runPlugin(pluginName, ba)
    resultsStr = fftPlugin.getResultStr()

    print('BINGO')
    print(resultsStr)

    sanpyWindow._fileListWidget._onLeftClick(1)
    sanpyWindow.myDetectionWidget.setAxis(0, 103.7)
    resultsStr = fftPlugin.getResultStr()

    sanpyWindow._fileListWidget._onLeftClick(2)
    sanpyWindow.myDetectionWidget.setAxis(16.4, 28.7)
    resultsStr = fftPlugin.getResultStr()

    print('BINGO')
    print(resultsStr)

def testFFT(sanpyWindow):
    sanpyWindow._fileListWidget._onLeftClick(1)
    #sanpyWindow.myDetectionWidget.setAxis(2.1,156.8)

    ba = sanpyWindow.get_bAnalysis()
    pluginName = 'FFT'
    fftPlugin = sanpyWindow.myPlugins.runPlugin(pluginName, ba)

def main():
    """Main entry point for the SanPy desktop app.
    
    Configured in setup.py
    """
    logger.info(f'Starting sanpy_app.py in __main__()')

    date_time_str = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    logger.info(f'    {date_time_str}')

    logger.info(f'    Python version is {platform.python_version()}')
    logger.info(f'    PyQt version is {QtCore.QT_VERSION_STR}')

    bundle_dir = sanpy._util.getBundledDir()
    logger.info(f'    bundle_dir is "{bundle_dir}"')

    _logFilePath = sanpy.sanpyLogger.getLoggerFile()
    logger.info(f'    logging to file {_logFilePath}')

    # for qdarkstyle
    #os.environ['PYQTGRAPH_QT_LIB'] = 'PyQt5'

    app = QtWidgets.QApplication(sys.argv)

    app.setQuitOnLastWindowClosed(True)

    # for manuscript we need to allow user to set light/dark theme
    # was this
    # v1
    #app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api=os.environ['PYQTGRAPH_QT_LIB']))
    # v2
    qdarktheme.setup_theme()

    #appIconPath = os.path.join(bundle_dir, 'interface/icons/sanpy_transparent.png')
    appIconPath = pathlib.Path(bundle_dir) / 'interface' / 'icons' / 'sanpy_transparent.png'
    appIconPathStr = str(appIconPath)
    #logger.info(f'appIconPath is "{appIconPath}"')
    if os.path.isfile(appIconPathStr):
        logger.info(f'    setting app window icon with: "{appIconPath}"')
        app.setWindowIcon(QtGui.QIcon(appIconPathStr))
    else:
        logger.warning(f'    Did not find appIconPath: {appIconPathStr}')

    w = SanPyWindow()
    w.show()

    w.raise_()  # bring to front, raise is a python keyword
    w.activateWindow()  # bring to front

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



"""
Test bAnalysis.py

Run this file with

```
python -m unittest sanpy/tests/test_analysis.py
```
"""

import os, shutil, tempfile
import unittest

import sanpy

import logging
from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__, level=logging.DEBUG)

class Test_Analysis(unittest.TestCase):
    # this patter is wierd to me?
    path = 'data/19114001.abf'
    # path = 'data/19114001.csv'
    #path = 'data/rosie-kymograph.tif'
    ba = sanpy.bAnalysis(path)
    expectedNumSpikes = 103
    expectedNumErrors = 69

    def test_0_load(self):
        logger.info('RUNNING')

        self.assertFalse(self.ba.loadError)
        self.assertIsNotNone(self.ba.fileLoader.sweepX)
        self.assertIsNotNone(self.ba.fileLoader.sweepY)
        self.assertEqual(len(self.ba.fileLoader.sweepX), len(self.ba.fileLoader.sweepY))

    def test_1_detect(self):
        logger.info('RUNNING')
        # grab detection parameters
        # see: sanpy.bDetection.detectionPresets.cakymograph
        bd = sanpy.bDetection()  # gets default
        dDict = bd.getDetectionDict('SA Node')
        # detect
        self.ba.spikeDetect(dDict)

        self.assertEqual(self.ba.numSpikes, self.expectedNumSpikes) # expecting 102 spikes
        self.assertEqual(len(self.ba.dfError), self.expectedNumErrors) # expecting 102 spikes

    def test_2_stats(self):
        logger.info('RUNNING')
        thresholdSec = self.ba.getStat('thresholdSec')

        self.assertEqual(len(thresholdSec), self.expectedNumSpikes) # expecting 102 spikes

if __name__ == '__main__':
    unittest.main()



"""
Author: RObert H Cudmore
Date: 20210811

Purpose:
	Provide functions to generate stimulation patterns and save them as Axon-Text-Files (atf).

Usage:
	coming soon.

Notes:
	Opening atf files in pClamp is slow, you want to open abf files instead.
	To acheive this, open the generated atf in Clampfit and save it as abf.

See:
	https://github.com/swharden/pyABF/tree/master/docs/advanced/creating-waveforms

"""

import sys, math
from math import exp
import numpy as np
import scipy.signal
import matplotlib.pyplot as plt

import sanpy

def getAtfHeader():
	ATF_HEADER = """
	ATF	1.0
	8	2
	"AcquisitionMode=Episodic Stimulation"
	"Comment="
	"YTop=2000"
	"YBottom=-2000"
	"SyncTimeUnits=20"
	"SweepStartTimesMS=0.000"
	"SignalsExported=IN 0"
	"Signals="	"IN 0"
	"Time (s)"	"Trace #1"
	""".strip()

	return ATF_HEADER

def padData(data, fs=10000, preSec=1, postSec=1, autoPad=True):
	"""
	All stimuli sent to pClamp need to be padded with zeros before/after the stimulus.

	pClamp something like zero(ing) out the first 1/16th duration of a stimulus.

	Note:
		- ClampEx cuts off the first 1/64th of a sweep
			TODO: Make sure our preSec of 1 second is LONGER than 1/64th of the sweep length !!!
		- Need to check this for V-Clamp protocols!!!

	Args:
		data (np.ndarray):
		fs (int): samlpes-per-unit-time, 10000 for 10 kHz
		preSec (int):
		postSec (int):
		autoPad (bool): If True will pad before/after data (1/16)*2 durations with zeros
	"""
	dataSeconds = len(data) / fs
	oneSixteenthSeconds = dataSeconds / 16
	if autoPad:
		preSec = oneSixteenthSeconds  # * 2
		postSec = oneSixteenthSeconds  # * 2
	# check that preSec is larger than 1/16 of total duration in data
	elif preSec < oneSixteenthSeconds:
		print(f'ERROR padData(): preSec is {preSec} seconds but needs to be more than 1/16th of data seconds {dataSeconds}.')
		print(f'  Make preSec > {oneSixteenthSeconds}')

	prePnts = preSec * fs  # seconds * samples-per-second
	postPnts = postSec * fs

	prePnts = math.ceil(prePnts)  # ceil rather than round for fear we lose 1 sampling point
	postPnts = math.ceil(postPnts)

	newLen_pnts = prePnts + len(data) + postPnts
	newData = np.zeros(newLen_pnts)
	newData[prePnts:prePnts+len(data)] = data

	return newData

def saveAtf(data, fileName="output.atf", fs=10000):
	"""Save a stimulus waveform array as an ATF 1.0 file."""
	out = getAtfHeader()

	# make a second sweep with noise (See if pyAbf can load it)
	#myNoise = np.random.normal(scale=np.sqrt(5), size=data.shape)

	for i,val in enumerate(data):
		#noiseVal = myNoise[i]
		# TODO: Convert to f'' format
		#out += '\n%.05f\t%.05f'%(i/fs,val)
		out += '\n%.05f\t%.05f'%(i/fs, val)
	with open(fileName,'w') as f:
		f.write(out)
		print(f'Saved: "{fileName}"')
	return True

def _loadAtf(path):
	"""
	Load our saved atf file. Primarily for testing the code.
	"""
	import pyabf

	# load
	abf = pyabf.ATF(path)
	print('_loadAtf()', abf)

	# plot
	data = abf.sweepY
	#plotData(data)
	return data

def plotData(data, fs=10000):
	"""
	"""
	# make a time axies
	t = np.arange(len(data)) / fs

	fig = plt.figure()
	ax1 = fig.add_subplot(1,1,1) #
	ax1.plot(t, data)
	ax1.set_xlabel('Time (s)')
	ax1.set_ylabel('pA') # assuming we are working in I-Clamp
	plt.show()

def _makeSin(durSec=10, amp=10, freqHz=10, fs=10000):
	"""
	Create a Sin wave.

	Args:
		durSec (float): Duration of stimulus
		amp (float): Amplitude of sin
		freqHz (float):
		fs (int): Sampling frequency in 'samples per unit time', 10000 for 10 kHz

	See: https://github.com/swharden/pyABF/tree/master/docs/advanced/creating-waveforms
	"""
	Xs = np.arange(fs*durSec) / fs  # x/time axis, we should always be able to create this from (dur, fs)
	data = amp * np.sin(2*np.pi*(Xs)*freqHz)

	return data

def _makeChirp(durSec=10, amp=10, maxFreqHz=10, fs=10000):
	"""
	Create a Sin wave sweep, a.k.a. a 'chirp'.

	Args:
		durSec (float): Duration of stimulus
		amp (float): Amplitude of sin
		maxFreqHz (float): Max frequency, chirp will go from/to (0, maxFreq_hz) Hz
		fs (int): Sampling frequency in 'samples per unit time', 10000 for 10 kHz
		autoPad (bool):

	See: https://github.com/swharden/pyABF/tree/master/docs/advanced/creating-waveforms
	"""
	timeScale = maxFreqHz / durSec / 2
	Xs = np.arange(fs*durSec) / fs  # x/time axis, we should always be able to create this from (dur, fs)
	data = amp * np.sin(2*np.pi*(Xs**2)*timeScale)

	#
	return data

def _makeNoise(durSec, noiseAmp, fs=10000):
	Xs = np.arange(fs*durSec) / fs  # x/time axis, we should always be able to create this from (dur, fs)
	#data = np.random.normal(scale=np.sqrt(noiseAmp), size=Xs.shape)
	data = np.random.normal(scale=noiseAmp, size=Xs.shape)
	return data

def _addNoise(data, noiseAmp):
	"""Add normal (gaussian) noise to existing data.

	Args:
		noiseAmp (float): After sqrt(noiseAmp) specifies the STD of noise.
	"""
	#data += np.random.normal(scale=np.sqrt(noiseAmp), size=data.shape)
	data += np.random.normal(scale=noiseAmp, size=data.shape)
	return data

def makeStim(type, sweepDurSec=10, startStimSec=2.0, durSec=5.0, yStimOffset=0,
					amp=10, freq=10, fs=10000,
					noiseAmp=1, rectify=True, autoPad=False, autoSave=True):
	"""
	Make a different number of stim

	Args:
		durSec (float): Duration of stim
		rectify (bool): If tru then stim<0 = 0
	"""

	# make the stim of duration durSec (then embed into longer sweepDurSec)
	if type == 'Noise':
		data = _makeNoise(durSec, noiseAmp, fs)
	elif type == 'Chirp':
		data = _makeChirp(durSec=durSec, amp=amp, maxFreqHz=freq, fs=fs)
	elif type == 'Sin':
		data = _makeSin(durSec=durSec, amp=amp, freqHz=freq, fs=fs)
	elif type == 'Epsp train':
		tmpSpikeTrain, data = getSpikeTrain(durSec=durSec, fs=10000, spikeFreq=freq, amp=amp, noiseAmp=0, baseLineShift=0)
	elif type == 'Integrate and Fire':
		data = integrateAndFire(durSec=durSec, freqHz=freq, fs=10000)
	# TODO: TOO SLOW TO BE USEFUL !!!
	elif type == 'Stochastic HH':
		from sanpy.models.myStochHH import myRun2
		data = sanpy.models.myStochHH.myRun2()
	else:
		print(f'makeStim() type "{type}" not understood.')
		return

	if noiseAmp > 0 and type != 'Noise':
		data = _addNoise(data, noiseAmp=noiseAmp)

	# always pad
	if autoPad:
		data = padData(data, autoPad=autoPad)

	if rectify:
		data[data<0] = 0

	# always add offset, if 0 will have no effect
	data += yStimOffset

	#
	# insert stim (data) into final sweep (sweepData)
	Xs = np.arange(fs*sweepDurSec) / fs  # x/time axis, we should always be able to create this from (dur, fs)
	sweepData = np.zeros(shape=Xs.shape)
	xStartPnt = int(fs * startStimSec)
	xStopPnt = xStartPnt + len(data)
	if xStopPnt > len(sweepData):
		xStopPnt = len(sweepData)
	sweepData[xStartPnt:xStopPnt] = data

	fileName = ''
	if autoSave:
		fileName = f'{type}_dur{durSec}_freq{freq}.atf'
		saveAtf(sweepData, fileName=fileName)

	#
	if autoSave:
		return sweepData, fileName
	else:
		return sweepData

def getKernel(type='sumExp', amp=5, tau1=30, tau2=70):
	"""Get a kernel for convolution with a spike train."""
	N = 500  # pnts
	t = [x for x in range(N)]
	y = t

	if type == 'sumExp':
		for i in t:
			y[i] = -amp * ( exp(-t[i]/tau1) - (exp(-t[i]/tau2)) )
	#
	return y

def getSpikeTrain(durSec=1, fs=10000, spikeFreq=3, amp=10, noiseAmp=10, baseLineShift=-60):
	"""Get a spike train at given frequency.

	Arguments:
		durSec (int): Total number of seconds
		fs (int): Sampling frequency, 10000 for 10 kH
		spikeFreq (int): Frequency of events in spike train, e.g. simulated EPSPs
		amp (float): Amplitude of sum exponential kernel (getKernel)
	"""
	n = int(durSec * fs) # total number of samples
	numSpikes = int(durSec * spikeFreq)
	spikeTrain = np.zeros(n)
	start = fs / spikeFreq
	spikeTimes = np.linspace(start, n, numSpikes, endpoint=False)
	for idx, spike in enumerate(spikeTimes):
		#print(idx, spike)
		spike = int(spike)
		spikeTrain[spike] = 1

	expKernel = getKernel(amp=amp)
	epspTrain = scipy.signal.convolve(spikeTrain, expKernel, mode='same')

	# shift to -60 mV
	epspTrain -= baseLineShift

	# add noise
	if noiseAmp == 0:
		pass
	else:
		noise_power = 0.001 * fs / noiseAmp
		epspTrain += np.random.normal(scale=np.sqrt(noise_power), size=epspTrain.shape)

	#
	#t = np.linspace(0, durSec, n, endpoint=True)

	#
	return spikeTrain, epspTrain


def integrateAndFire(durSec, freqHz, fs):
	"""
	See: https://mark-kramer.github.io/Case-Studies-Python/IF.html
	"""
	print('integrateAndFire()')
	#durSec = 50
	#fs = 10000
	dt = 1 / fs
	durPnts = (1/dt) * durSec
	durPnts = int(durPnts)
	print('  durPnts:', durPnts)

	# constant current
	constantAmp = 0.7
	I = np.zeros(durPnts) + constantAmp
	print(f'  constantAmp:{constantAmp} I:{I.shape}')
	# plus sin
	sinAmp = 0.5
	sinFreq = freqHz
	rmsMult = 1/np.sqrt(2)
	sinRms = sinAmp * rmsMult
	# want sinRms to equal constantAMp
	sinAmp = constantAmp / rmsMult
	iSin = _makeSin(durSec=durSec, amp=sinAmp, freqHz=sinFreq, fs=fs)
	print(f'  sinAmp:{sinAmp} sinRms:{sinRms} iSin:{iSin.shape}')
	I += iSin
	# Pluse noise
	noiseAmp = 1
	I += np.random.normal(scale=noiseAmp, size=I.shape)

	#I = 2					  # constant current
	C = 1						  # capacitance
	Vth = 1;					 # voltage threshold.
	Vreset = 0;				  # reset voltage.
	V = np.zeros(durPnts)		# Initialize V.
	V[0] = 0.2;					# Set the initial condition.

	spikeTimesSeconds = []
	for k in range(1,durPnts-1):	   # March forward in time,
		V[k+1] = V[k] + dt*(I[k]/C) # Update the voltage,
		if V[k+1] > Vth:		 # ... and check if the voltage exceeds the threshold.
			V[k+1] = Vreset
			spikeTimesSeconds.append(k*dt)

	print(f'  numSpikes:{len(spikeTimesSeconds)}')
	spikeTimeAmp = [Vth] * len(spikeTimesSeconds)

	t = np.arange(0,len(V))*dt	  # Define the time axis.

	'''
	fig, axs = plt.subplots(3, 1, sharex=True)
	axs[0].plot(t,I, '-k', linewidth=1)
	axs[1].plot(t,V, '-k', linewidth=1)
	axs[1].plot(spikeTimesSeconds, spikeTimeAmp, 'or')

	isi_sec = np.diff(spikeTimesSeconds)
	spikeFreq = [1/isi for isi in isi_sec]
	axs[2].plot(spikeTimesSeconds[1:], spikeFreq, 'o')

	axs[0].set_ylabel('Current')
	axs[1].set_ylabel('Vm')

	axs[2].set_xlabel('Time (s)')
	axs[2].set_ylabel('Spike Frequency (Hz)')

	#
	plt.show()
	'''

	return V

if __name__ == '__main__':
	amp = 10
	durSec = 20
	fs = 10000

	data = makeStim('sin')
	data = makeStim('sin + noise')
	data = makeStim('chirp')

	integrateAndFire()

	'''
	plotData(data)

	loadedData = _loadAtf(fileName)
	plotData(loadedData)
	'''



from typing import Union, Dict, List, Tuple
import numpy as np
#import pandas as pd

import tifffile

import sanpy
from sanpy.fileloaders.fileLoader_base import fileLoader_base

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class fileLoader_tif(fileLoader_base):
    loadFileType = 'tif'
    
    # @property
    # def loadFileType(self):
    #     return 'tif'

    def loadFile(self):
        # assuming pixels x line scan like (519, 10000)
        self._tif = tifffile.imread(self.filepath)

        dt = 0.001  # sec

        # using 'reshape(-1,1)' to convert shape from (n,) to (n,1)
        sweepX = np.arange(0, self._tif.shape[1]).reshape(-1,1)
        sweepY = np.sum(self._tif, axis=0).reshape(-1,1)
    
        self._dt = 1
        self._dy = 1

        self.setLoadedData(
                sweepX = sweepX,
                sweepY = sweepY,
        )

    #
    # need to pull/merge code from xxx
    # bAbfText._abfFromLineScanTif()

    def resetKymographRect(self):
        defaultRect = self._abf.resetRoi()
        self._updateTifRoi(defaultRect)
        return defaultRect

    def getKymographRect(self):
        if self.isKymograph():
            return self._abf.getTifRoi()
        else:
            return None

    def getKymographBackgroundRect(self):
        if self.isKymograph():
            return self._abf.getTifRoiBackground()
        else:
            return None

    def _updateTifRoi(self, theRect=[]):
        """
        Update the kymograph ROI
        """
        self._abf.updateTifRoi(theRect)

        self._sweepX[:, 0] = self._abf.sweepX
        self._sweepY[:, 0] = self._abf.sweepY




from setuptools import setup, find_packages

VERSION = "0.1.6"

setup(
    name='sanpy-ephys',  # thepackage name (on PyPi), still use 'import sanpy'
    version=VERSION,
    description='Whole cell current clamp analysis.',
    url='http://github.com/cudmore/SanPy',
    author='Robert H Cudmore',
    author_email='robert.cudmore@gmail.com',
    license='GNU General Public License, Version 3',
    # this is CRITICAL to import submodule like sanpy.userAnalysis
    packages=find_packages(include=['sanpy', 'sanpy.*', 'sanpy.interface', 'sanpy.fileloaders']),
    install_requires=[
        'numpy',
        'pandas',
        'scipy',
        'pyabf',
        'tifffile',
        #'XlsxWriter',
        #'xlrd', #  for loading excel files in examples/reanalyze.py
        #'openpyxl',
        'matplotlib',
        'mplcursors',
        'seaborn',
        'requests', #  to load from the cloud (for now github)
        'tables',  # aka pytable for hdf5.
        # used for line profile in kym analysis
        # 0.20.0 introduces pyinstaller bug because of lazy import
        'scikit-image==0.19.3', 
        'h5py',
    ],
    extras_require={
        'gui': [
            'qtpy',
            #'pyqtgraph==0.12.4',
            'pyqtgraph',
            #'qdarkstyle',  # v1
            'pyqtdarktheme',  # switched to this mar 2023
            #'PyQt5==5.15.6',  #
            'PyQt5',  #
        ],
        'dev': [
            'mkdocs',
            'mkdocs-material',
            'mkdocs-jupyter',
            'mkdocstrings',
            'tornado', # needed for pyinstaller
            'pyinstaller',
            'ipython',
            'tox',
            'pytest',
            'pytest-cov',
            'pytest-qt',
            'flake8',
            'jupyter',
            'pooch'  # what is this for?
        ],
        #"test": ["pytest", "pytest-cov", "scikit-image", "pooch"],
        'test': [
            'pytest',
            'pytest-cov',
            #'pytest-qt',
            'flake8',
        ]
    },
    python_requires=">=3.7",
    entry_points={
        'console_scripts': [
            'sanpy=sanpy.interface.sanpy_app:main',
        ]
    },
)



#from . import *

from .sanpyPlugin import *

# TODO: Fix this so I do not need to add each
from .plotRecording import plotRecording
from .sanpyLog import sanpyLog

from .basePlotTool import basePlotTool
from .plotTool import plotTool
from .plotToolPool import plotToolPool

from .plotScatter import plotScatter
from .detectionErrors import detectionErrors
from .spikeClips import spikeClips
from .resultsTable import resultsTable
from .resultsTable2 import resultsTable2
from .analysisSummary import analysisSummary
from .exportTrace import exportTrace

from .fftPlugin import fftPlugin
from .stimGen import stimGen

from .detectionParams import detectionParams

from .kymographPlugin import kymographPlugin
#from .kymographPlugin2 import kymographPlugin2

from .setSpikeStat import SetSpikeStat

# eventually move this out of plotScatter
from .plotScatter import myStatListWidget

from .plotFi import plotFi

# TODO: make this just one line, so user can drop a bplugin inn and restart
from . import *



import pandas as pd

from PyQt5 import QtWidgets

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class analysisSummary(sanpyPlugin):
    """Plugin to display overview of analysis.

    Notes
    -----
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """
    myHumanName = 'Summary Analysis'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.df = None

        #self.pyqtWindow()  # makes self.mainWidget

        layout = QtWidgets.QVBoxLayout()
        self.myErrorTable = sanpy.interface.bErrorTable.errorTableView()
        self.myErrorTable.setSortingEnabled(False)

        layout.addWidget(self.myErrorTable)

        # mar 17
        # self.setLayout(layout)
        self.getVBoxLayout().addLayout(layout)

        self.replot()

    def replot(self):
        """Grab a report from sanpy.bExport.getSummary()
        Put it in a table.
        """
        logger.info('')

        if self.ba is None:
            return

        exportObject = sanpy.bExport(self.ba)
        theMin, theMax = self.getStartStop()
        dfSummary = exportObject.getSummary(theMin=theMin, theMax=theMax)
        if dfSummary is not None:
            errorReportModel = sanpy.interface.bFileTable.pandasModel(dfSummary)
            self.myErrorTable.setModel(errorReportModel)
        else:
            # no spikes
            logger.info(f'No spikes to report')
            dfSummary = pd.DataFrame(columns=['', '', '', '', ''])
            errorReportModel = sanpy.interface.bFileTable.pandasModel(dfSummary)
            self.myErrorTable.setModel(errorReportModel)

        # hard-cde column widths
        self.myErrorTable.setColumnWidth(0,300)
        self.myErrorTable.setColumnWidth(1,300)

        #
        self.df = dfSummary

    def setAxis(self):
        # inherited
        # regenerate summary with new range
        self.replot()

    def copyToClipboard(self):
        if self.df is not None:
            print(self.df.head())
            self.df.to_clipboard(sep='\t', index=False)  # index=False so we do not duplicate 1s column
            logger.info('Copied to clipboard')

if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication([])

    # load and analyze sample data
    path = '/home/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()

    # open a plugin
    sa = analysisSummary(ba=ba, startStop=None)
    sa.show()

    sys.exit(app.exec_())




"""sanpy fft backend.

TODO:
	- get rid of (spikeDetect, _getHalfWidths, _throwOutRefractory)
		- these are used to spike detect model data
		- try and use main bAnalysis_
"""

import math, time, os
from math import exp
from functools import partial
from tokenize import blank_re

import numpy as np
import pandas as pd

from scipy.signal import lfilter, freqz
import scipy.signal

import matplotlib as mpl
import matplotlib.pyplot as plt

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def butter_sos(cutOff, fs, order=50, passType='lowpass'):
	"""
	passType (str): ('lowpass', 'highpass')
	"""

	'''
	nyq = fs * 0.5
	normal_cutoff = cutOff / nyq
	'''

	logger.info(f'Making butter file with order:{order} cutOff:{cutOff} passType:{passType} fs:{fs}')

	#sos = butter(order, normal_cutoff, btype='low', analog=False, output='sos')
	#sos = scipy.signal.butter(N=order, Wn=normal_cutoff, btype=passType, analog=False, fs=fs, output='sos')
	sos = scipy.signal.butter(N=order, Wn=cutOff, btype=passType, analog=False, fs=fs, output='sos')
	return sos

def spikeDetect(t, dataFiltered, dataThreshold2, startPoint=None, stopPoint=None):
	verbose = False
	#dataThreshold2 = -59.81

	g_backupFraction = 0.3  # 0.5 or 0.3

	logger.info(f'dataThreshold2:{dataThreshold2} startPoint:{startPoint} stopPoint:{stopPoint}')

	#
	# detect spikes, dataThreshold2 changes as we zoom
	Is = np.where(dataFiltered>dataThreshold2)[0]
	Is=np.concatenate(([0],Is))
	Ds=Is[:-1]-Is[1:]+1
	spikePoints = Is[np.where(Ds)[0]+1]
	spikeSeconds = t[spikePoints]

	#
	# Throw out spikes that are within refractory (ms) of previous
	spikePoints = _throwOutRefractory(spikePoints, refractory_ms=100)
	spikeSeconds = t[spikePoints]

	#
	# remove spikes outside of start/stop
	goodSpikes = []
	if startPoint is not None and stopPoint is not None:
		if verbose:
			print(f'  Removing spikes outside point range {startPoint} ... {stopPoint}')
		#print(f'    {t[startPoint]} ... {t[stopPoint]} (s)')
		for idx, spikePoint in enumerate(spikePoints):
			if spikePoint<startPoint or spikePoint>stopPoint:
				pass
			else:
				goodSpikes.append(spikePoint)
				if verbose:
					print(f'    appending spike {idx} {t[spikePoint]}(s)')
	#
	spikePoints = goodSpikes
	spikeSeconds = t[spikePoints]

	#
	# only accept rising phase
	windowPoints = 40 # assuming 10 kHz -->> 4 ms
	goodSpikes = []
	for idx, spikePoint in enumerate(spikePoints):
		preClip = dataFiltered[spikePoint-windowPoints:spikePoint-1]
		postClip = dataFiltered[spikePoint+1:spikePoint+windowPoints]
		preMean = np.nanmean(preClip)
		postMean = np.nanmean(postClip)
		if preMean < postMean:
			goodSpikes.append(spikePoint)
		else:
			pass
			#print(f'  rejected spike {idx}, at point {spikePoint}')
	#
	spikePoints = goodSpikes
	spikeSeconds = t[spikePoints]

	#
	# throw out above take-off-potential of APs (-30 mV)
	apTakeOff = -30
	goodSpikes = []
	for idx, spikePoint in enumerate(spikePoints):
		if dataFiltered[spikePoint] < apTakeOff:
			goodSpikes.append(spikePoint)
	#
	spikePoints = goodSpikes
	spikeSeconds = t[spikePoints]

	#
	# find peak for eack spike
	# at 500 pnts, assuming spikes are slower than 20 Hz
	peakWindow_pnts = 500
	peakPoints = []
	peakVals = []
	for idx, spikePoint in enumerate(spikePoints):
		peakPnt = np.argmax(dataFiltered[spikePoint:spikePoint+peakWindow_pnts])
		peakPnt += spikePoint
		peakVal = np.max(dataFiltered[spikePoint:spikePoint+peakWindow_pnts])
		#
		peakPoints.append(peakPnt)
		peakVals.append(peakVal)

	#
	# backup each spike until pre/post is not changing
	# uses g_backupFraction to get percent of dv/dt at spike veruus each backup window
	backupWindow_pnts = 30
	maxSteps = 30  # 30 steps at 20 pnts (2ms) per step gives 60 ms
	backupSpikes = []
	for idx, spikePoint in enumerate(spikePoints):
		#print(f'=== backup spike:{idx} at {t[spikePoint]}(s)')
		foundBackupPoint = None
		for step in range(maxSteps):
			tmpPnt = spikePoint - (step*backupWindow_pnts)
			tmpSeconds = t[tmpPnt]
			preClip = dataFiltered[tmpPnt-1-backupWindow_pnts:tmpPnt-1]  # reversed
			postClip = dataFiltered[tmpPnt+1:tmpPnt+1+backupWindow_pnts]
			preMean = np.nanmean(preClip)
			postMean = np.nanmean(postClip)
			diffMean = postMean - preMean
			if step == 0:
				initialDiff = diffMean
			#print(f'  spike {idx} step:{step} tmpSeconds:{round(tmpSeconds,4)} initialDiff:{round(initialDiff,4)} diff:{round(diffMean,3)}')
			# if diffMean is 1/2 initial AP slope then accept
			if diffMean < (initialDiff * g_backupFraction):
				# stop
				if foundBackupPoint is None:
					foundBackupPoint = tmpPnt
				#break
		#
		if foundBackupPoint is not None:
			backupSpikes.append(foundBackupPoint)
		else:
			# needed to keep spike parity
			logger.warning(f'Did not find backupSpike for spike {idx} at {t[spikePoint]}(s)')
			backupSpikes.append(spikePoint)

	#
	# use backupSpikes (points) to get each spike amplitude
	spikeAmps = []
	for idx, backupSpike in enumerate(backupSpikes):
		if backupSpikes==np.nan or math.isnan(backupSpike):
			continue
		#print('backupSpike:', backupSpike)
		footVal = dataFiltered[backupSpike]
		peakVal = peakVals[idx]
		spikeAmp = peakVal - footVal
		spikeAmps.append(spikeAmp)

	# TODO: use foot and peak to get real half/width
	theseWidths=[10,20,50,80,90]
	window_ms=100
	spikeDictList = _getHalfWidths(t, dataFiltered,
						backupSpikes, peakPoints,
						theseWidths=theseWidths, window_ms=window_ms)
	#for idx, spikeDict in enumerate(spikeDictList):
	#	print(idx, spikeDict)

	#
	# get estimate of duration
	# for each spike, find next downward crossing (starting at peak)
	'''
	minDurationPoints = 10
	windowPoints = 1000  # 100 ms
	fallingPoints = []
	for idx, spikePoint in enumerate(spikePoints):
		peakPoint = peakPoints[idx]
		backupPoint = backupSpikes[idx]
		if backupPoint == np.nan or math.isnan(backupPoint):
			continue
		spikeSecond = spikeSeconds[idx]
		# Not informative because we are getting spikes from absolute threshold
		thisThreshold = dataFiltered[backupPoint]
		thisPeak = dataFiltered[peakPoint]
		halfHeight = thisThreshold + (thisPeak - thisThreshold) / 2
		startPoint = peakPoint #+ 50
		postClip = dataFiltered[startPoint:startPoint+windowPoints]
		tmpFallingPoints = np.where(postClip<halfHeight)[0]
		if len(tmpFallingPoints) > 0:
			fallingPoint = startPoint + tmpFallingPoints[0]
			# TODO: check if falling point is AFTER next spike
			duration = fallingPoint - spikePoint
			if duration > minDurationPoints:
				fallingPoints.append(fallingPoint)
			else:
				print(f'  reject spike {idx} at {spikeSecond} (s), duration is {duration} points, minDurationPoints:{minDurationPoints}')
				#fallingPoints.append(fallingPoint)
		else:
			print(f'    did not find falling pnt for spike {idx} at {spikeSecond}(s) point {spikePoint}, assume it is longer than windowPoints')
			pass
			#fallingPoints.append(np.nan)
	'''

	# TODO: Package all results into a dictionary
	spikeDictList = [{}] * len(spikePoints)
	for idx, spikePoint in enumerate(spikePoints):
		'''
		spikeSecond = spikeSeconds[idx]
		peakPoint = peakPoints[idx]
		peakVal = peakVals[idx]
		fallingPoint = fallingPoints[idx]  # get rid of this
		backupSpike = backupSpikes[idx]  # get rid of this
		spikeAmp = spikeAmps[idx]  # get rid of this
		'''
		spikeDictList[idx]['spikePoint'] = spikePoint
		spikeDictList[idx]['spikeSecond'] = spikeSeconds[idx]
		spikeDictList[idx]['peakPoint'] = peakPoints[idx]
		spikeDictList[idx]['peakVal'] = peakVals[idx]
		#spikeDictList[idx]['fallingPoint'] = fallingPoints[idx]
		spikeDictList[idx]['backupSpike'] = backupSpikes[idx]
		spikeDictList[idx]['spikeAmp'] = spikeAmps[idx]
	#
	spikeSeconds = t[spikePoints]
	#return spikeSeconds, spikePoints, peakPoints, peakVals, fallingPoints, backupSpikes, spikeAmps
	# removved fallingPoints
	return spikeDictList, spikeSeconds, spikePoints, peakPoints, peakVals, backupSpikes, spikeAmps

def _getHalfWidths(t, v, spikePoints, peakPoints, theseWidths=[10,20,50,80,90], window_ms=50):
	"""Get half widths.

	Args:
		t (ndarray): Time
		v (ndaray): Recording (usually curent clamp Vm)
		spikePoints (list of int): List of spike threshold crossings.
							Usually the back-up version
		peakPoints (list of int):
		theseWidths (list of int): Specifies full-width-half maximal to calculate
		window_ms (int): Number of ms to look after the peak for downward 1/2 height crossing

	Returns:
		List of dict, one elemennt per spike
	"""
	logger.info(f'theseWidths:{theseWidths} window_ms:{window_ms} spikePoints:{len(spikePoints)} peakPoints:{len(peakPoints)}')

	pointsPerMs = 10  # ToDo: generalize this
	window_pnts = window_ms * pointsPerMs

	spikeDictList = [{}] * len(spikePoints) # an empy list of dict with proper size
	for idx, spikePoint in enumerate(spikePoints):
		#print(f'Spike {idx} pnt:{spikePoint}')

		# each spike has a list of dict for each width result
		spikeDictList[idx]['widths'] = []

		# check each pre/post pnt is before next spike
		nextSpikePnt = None
		if idx < len(spikePoints)-1:
			nextSpikePnt = spikePoints[idx+1]

		peakPoint = peakPoints[idx]
		try:
			vSpike = v[spikePoint]
		except (IndexError) as e:
			logger.error(f'spikePoint:{spikePoint} {type(spikePoint)}')
		vPeak = v[peakPoint]
		height = vPeak-vSpike
		preStartPnt = spikePoint+1
		preClip = v[preStartPnt:peakPoint-1]
		postStartPnt = peakPoint+1
		postClip = v[postStartPnt:peakPoint+window_pnts]
		for width in theseWidths:
			thisHeight = vSpike + (height / width)  # search for a percent of height
			prePnt = np.where(preClip>=thisHeight)[0]
			if len(prePnt) > 0:
				prePnt = preStartPnt + prePnt[0]
			else:
				#print(f'  Error: Spike {idx} "prePnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
				prePnt = None
			postPnt = np.where(postClip<thisHeight)[0]
			if len(postPnt) > 0:
				postPnt = postStartPnt + postPnt[0]
			else:
				#print(f'  Error: Spike {idx} "postPnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
				postPnt = None
			widthMs = None
			if prePnt is not None and postPnt is not None:
				widthPnts = postPnt - prePnt
				widthMs = widthPnts / pointsPerMs
				#print(f'  width:{width} tPre:{t[prePnt]} tPost:{t[postPnt]} widthMs:{widthMs}')
				if nextSpikePnt is not None and prePnt >= nextSpikePnt:
					print(f'  Error: Spike {idx} widthMs:{widthMs} prePnt:{prePnt} is after nextSpikePnt:{nextSpikePnt}')
				if nextSpikePnt is not None and postPnt >= nextSpikePnt:
					print(f'  Error: Spike {idx} widthMs:{widthMs} postPnt:{postPnt} is after nextSpikePnt:{nextSpikePnt}')

			# put into dict
			widthDict = {
				'halfHeight': width,
				'risingPnt': prePnt,
				#'risingVal': defaultVal,
				'fallingPnt': postPnt,
				#'fallingVal': defaultVal,
				#'widthPnts': None,
				'widthMs': widthMs
			}
			spikeDictList[idx]['widths_' + str(width)] = widthMs
			spikeDictList[idx]['widths'].append(widthDict)
	#
	return spikeDictList

def _throwOutRefractory(spikePoints, refractory_ms=100):
	"""
	spikePoints: spike times to consider
	refractory_ms:
	"""
	dataPointsPerMs = 10

	before = len(spikePoints)

	# if there are doubles, throw-out the second one
	#refractory_ms = 20 #10 # remove spike [i] if it occurs within refractory_ms of spike [i-1]
	lastGood = 0 # first spike [0] will always be good, there is no spike [i-1]
	for i in range(len(spikePoints)):
		if i==0:
			# first spike is always good
			continue
		dPoints = spikePoints[i] - spikePoints[lastGood]
		if dPoints < dataPointsPerMs*refractory_ms:
			# remove spike time [i]
			spikePoints[i] = 0
		else:
			# spike time [i] was good
			lastGood = i
	# regenerate spikeTimes0 by throwing out any spike time that does not pass 'if spikeTime'
	# spikeTimes[i] that were set to 0 above (they were too close to the previous spike)
	# will not pass 'if spikeTime', as 'if 0' evaluates to False
	#if goodSpikeErrors is not None:
	#	goodSpikeErrors = [goodSpikeErrors[idx] for idx, spikeTime in enumerate(spikeTimes0) if spikeTime]
	spikePoints = [spikePoint for spikePoint in spikePoints if spikePoint]

	# TODO: put back in and log if detection ['verbose']
	after = len(spikePoints)
	logger.info(f'From {before} to {after} spikes with refractory_ms:{refractory_ms}')

	return spikePoints

def getKernel(type='sumExp', amp=5, tau1=30, tau2=70):
	"""Get a kernel for convolution with a spike train."""
	N = 500  # pnts
	t = [x for x in range(N)]
	y = t

	if type == 'sumExp':
		for i in t:
			y[i] = -amp * ( exp(-t[i]/tau1) - (exp(-t[i]/tau2)) )
	#
	return y

def getSpikeTrain(numSeconds=1, fs=10000, spikeFreq=3, amp=10, noiseAmp=10):
	"""Get a spike train at given frequency.

	Arguments:
		numSeconds (int): Total number of seconds
		fs (int): Sampling frequency, 10000 for 10 kH
		spikeFreq (int): Frequency of events in spike train, e.g. simulated EPSPs
		amp (float): Amplitude of sum exponential kernel (getKernel)
	"""
	n = int(numSeconds * fs) # total number of samples
	numSpikes = int(numSeconds * spikeFreq)
	spikeTrain = np.zeros(n)
	start = fs / spikeFreq
	spikeTimes = np.linspace(start, n, numSpikes, endpoint=False)
	for idx, spike in enumerate(spikeTimes):
		#print(idx, spike)
		spike = int(spike)
		spikeTrain[spike] = 1

	expKernel = getKernel(amp=amp)
	epspTrain = scipy.signal.convolve(spikeTrain, expKernel, mode='same')
	# shift to -60 mV
	epspTrain -= 60

	# add noise
	if noiseAmp == 0:
		pass
	else:
		#noise_power = 0.001 * fs / noiseAmp
		#epspTrain += np.random.normal(scale=np.sqrt(noise_power), size=epspTrain.shape)
		epspTrain += np.random.normal(scale=noiseAmp, size=epspTrain.shape)

	#
	t = np.linspace(0, numSeconds, n, endpoint=True)
	#
	return t, spikeTrain, epspTrain

class fftFromRecording():

	def __init__(self, path=None, ba=None,
					sweepNumber=0, startSec=None, stopSec=None,
					doPlot=True):
		"""fft backend.
		
		Args:
			path (str) if specified, load abf from path
			ba (bAnalysis) if specified use that
		"""
		
		if ba is not None:
			self.ba = ba
		elif path is not None:
			self.ba = sanpy.bAnalysis(path)
		else:
			logger.error(f'Expecting either a path or existing bAnalysis object')

		#self.ba.setSweep(sweepNumber)
		self.sweepNumber = sweepNumber

		if startSec is None or stopSec is None:
			self.startSec = 0
			self.stopSec = self.ba.recordingDur
		else:
			self.startSec = startSec
			self.stopSec = stopSec

		self.plotRawAxis = True

		self.doButterFilter = True
		self.butterOrder = 70
		self.butterCutoff = [0.7, 10]  # [low freq, high freq] for bandpass

		self._resultsDictList = []
		self._resultsDictList2 = []
		self._resultStr = ''

		self.isModel = False

		self._store_ba = None  # allow switching between model and self.ba

		if self.ba is not None:
			self.fs = self.ba.recordingFrequency * 1000
		else:
			self.fs = None
		self.psdWindowStr = 'Hanning'  # mpl.mlab.window_hanning

		# fft Freq (Hz) resolution is fs/nfft --> resolution 0.2 Hz = 10000/50000
		self.nfft = 50000 #512 * 100

		self.signalHz = None  # assign when using fake data
		self.maxPlotHz = 10  # limit x-axis frequenccy

		self.medianFilterPnts = 0

		self.dataLine = None
		self.dataFilteredLine = None
		self.spikesLine = None
		self.peaksLine = None
		self.dataMeanLine = None
		self.thresholdLine2 = None
		self.thresholdLine3 = None

		#self.fig = mpl.figure.Figure(constrained_layout=True)
		self.fig = plt.figure(constrained_layout=True)

		if self.plotRawAxis:
			numRow = 3
		else:
			numRow = 2

		gs = self.fig.add_gridspec(numRow, 3)

		if self.plotRawAxis:
			self.rawAxes = self.fig.add_subplot(gs[0,:])
			self.rawZoomAxes = self.fig.add_subplot(gs[1,:])
			# 3rd row has 3 subplots
			self.psdAxes = self.fig.add_subplot(gs[2,0])
			self.spectrogramAxes = self.fig.add_subplot(gs[2,-2])
		else:
			self.rawAxes = None
			self.rawZoomAxes = self.fig.add_subplot(gs[0,:])
			# 3rd row has 3 subplots
			self.psdAxes = self.fig.add_subplot(gs[1,0])
			self.spectrogramAxes = self.fig.figure.add_subplot(gs[1,-2])

		if doPlot:
			self.replot_fft2(sweepNumber=self.sweepNumber)

	def getFigFileName(self):
		filename = 	self.ba.getFileName()
		filename = os.path.splitext(filename)[0]
		sweep = self.ba.currentSweep
		figFileName = f'{filename}_sweep_{sweep}.png'
		return figFileName

	def saveFig(self, folderPath=None):
		"""Save the current figure.
		"""

		# TODO: if folderPath then check it exists

		figFileName = self.getFigFileName()
		figFilePath = os.path.join(folderPath, figFileName)
		
		logger.info(f'Saving figure to: {figFilePath}')
		
		self.fig.savefig(figFilePath, dpi=300)


	def getStartStop(self):
		#logger.warning('remember to set start/stop')
		#startSec = 0
		#stopSec = self.ba.recordingDur
		return self.startSec, self.stopSec

	def linearDetrend(self, x):
		#print('myDetrend() x:', x.shape)
		y = plt.mlab.detrend_linear(x)
		#y = plt.mlab.detrend_mean(y)
		return y

	def replot_fft2(self, sweepNumber=0):

		start = time.time()

		if self.ba is None:
			return

		self.ba.setSweep(sweepNumber)

		#if self.sweepNumber == 'All':
		#	logger.warning(f'fft plugin can only show one sweep, received sweepNumber:{self.sweepNumber}')
		#	return

		#logger.info(f'using ba: {self.ba}')
		startSec, stopSec = self.getStartStop()
		if startSec is None or stopSec is None or math.isnan(startSec) or math.isnan(stopSec):
			logger.info(f'Resetting start/stop seconds to max')
			startSec = 0
			stopSec = self.ba.recordingDur  #self.ba.sweepX()[-1]
		logger.info(f'Using start(s):{startSec} stop(s):{stopSec}')
		self.lastLeft = round(startSec*1000*self.ba.dataPointsPerMs)
		self.lastRight = round(stopSec*1000*self.ba.dataPointsPerMs)

		leftPoint = self.lastLeft
		rightPoint = self.lastRight

		#sweepY = self.getSweep('y')
		sweepY = self.ba.sweepY
		if leftPoint<0:
			leftPoint = 0
		if rightPoint > sweepY.shape[0]:
			rightPoint = sweepY.shape[0]

		y = sweepY[leftPoint:rightPoint]

		medianPnts = self.medianFilterPnts  # 50
		if medianPnts > 0:
			logger.info(f'  Median filter with pnts={medianPnts}')
			yFiltered = scipy.ndimage.median_filter(y, medianPnts)
		else:
			logger.info('  No median filter')
			yFiltered = y

		#logger.info(f'Fetching sweepX with sweepNumber: {self.sweepNumber}')

		#sweepX = self.getSweep('x')
		sweepX = self.ba.fileLoader.sweepX

		#logger.info(f'  sweepX: {sweepX.shape}')
		#logger.info(f'  leftSec:{sweepX[leftPoint]} rightSec:{sweepX[rightPoint-1]}')
		t = sweepX[leftPoint:rightPoint]

		dt = t[1] - t[0] #0.0001
		fs = round(1/dt)  # samples per second
		nfft = self.nfft  # The number of data points used in each block for the FFT
		#print(f'  fs:{fs} nfft:{nfft}')

		# 20220329 hijacking raw plot to show stim
		if self.plotRawAxis:
			sweepC = self.ba.sweepC
			
			self.rawAxes.clear()
			#self.rawAxes.plot(sweepX, sweepY, '-', linewidth=1)
			self.rawAxes.plot(sweepX, sweepC, '-', linewidth=1)
			# draw a rectangle to show left/right time selecction
			'''
			yMin = np.nanmin(sweepY)
			yMax = np.nanmax(sweepY)
			xRect = startSec
			yRect = yMin
			wRect = stopSec-startSec
			hRect = yMax-yMin
			rect1 = mpl.patches.Rectangle((xRect,yRect), wRect, hRect, color='gray')
			self.rawAxes.add_patch(rect1)
			'''
			#self.rawAxes.set_xlim([sweepX[0], sweepX[-1]])
			self.rawAxes.set_xlim([self.startSec, self.stopSec])

		#
		# replot the clip +/- std we analyzed, after detrend
		yDetrend = self.linearDetrend(yFiltered)

		#
		if self.doButterFilter:
			#self.butterCutoff = 10  # Hz
			#self.butterOrder = 70
			#self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='lowpass')
			#self.butterCutoff = [0.7, 10]
			#logger.info(f'  butterOrder:{self.butterOrder} butterCutoff:{self.butterCutoff}')
			self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='bandpass')
			# filtfilt should remove phase delay in time. A forward-backward digital filter using cascaded second-order sections.
			yFiltered_Butter = scipy.signal.sosfiltfilt(self.sos, yDetrend, axis=0)
			#print('yDetrend:', yDetrend.shape, np.nanmin(yDetrend), np.nanmax(yDetrend))
			#print('yFiltered_Butter:', yFiltered_Butter.shape, np.nanmin(yFiltered_Butter), np.nanmax(yFiltered_Butter))

			# we use this for remaining
			yFiltered = yFiltered_Butter

		dataMean = np.nanmean(yDetrend)
		dataStd = np.nanstd(yDetrend)
		dataThreshold2 = dataMean + (2 * dataStd)
		dataThreshold3 = dataMean + (3 * dataStd)

		self.rawZoomAxes.clear()
		self.rawZoomAxes.plot(t, yDetrend, '-', linewidth=1)
		self.rawZoomAxes.axhline(dataMean, marker='', color='r', linestyle='--', linewidth=0.5)
		self.rawZoomAxes.axhline(dataThreshold2, marker='', color='r', linestyle='--', linewidth=0.5)
		self.rawZoomAxes.axhline(dataThreshold3, marker='', color='r', linestyle='--', linewidth=0.5)
		self.rawZoomAxes.set_xlim([t[0], t[-1]])
		#
		if self.doButterFilter:
			self.rawZoomAxes.plot(t, yFiltered, '-r', linewidth=1)

		# save yDetrend, yFiltered as csv
		'''
		print('=== FOR FERNANDO')
		import pandas as pd
		print('  t:', t.shape)
		print('  yDetrend:', yDetrend.shape)
		print('  yFiltered:', yFiltered.shape)
		tmpDf = pd.DataFrame(columns=['s', 'yDetrend', 'yFiltered'])
		tmpDf['s'] = t
		tmpDf['yDetrend'] = yDetrend
		tmpDf['yFiltered'] = yFiltered
		print(tmpDf.head())
		dfFile = 'fft-20200707-0000-16sec-31sec.csv'
		print('saving dfFile:', dfFile)
		tmpDf.to_csv(dfFile, index=False)
		print('=== END')
		'''

		# we will still scale a freq plots to [0, self.maxPlotHz]
		if self.doButterFilter:
			minPlotFreq = self.butterCutoff[0]
			maxPlotFreq = self.butterCutoff[1]
		else:
			minPlotFreq = 0
			maxPlotFreq = self.maxPlotHz #15

		#
		# spectrogram
		#self.fftAxes.clear()
		nfft2 = int(nfft / 4)
		specSpectrum, specFreqs, spec_t, specIm = self.spectrogramAxes.specgram(yFiltered, 
											NFFT=nfft2, Fs=fs, detrend=self.linearDetrend)
		'''
		print('=== specgram')
		print('  yFiltered:', yFiltered.shape)
		print('  specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
		print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
		print('  specFreqs:', specFreqs.shape)
		print('  spec_t:', spec_t.shape)
		print('  specIm:', type(specIm))
		'''
		mask = (specFreqs >= minPlotFreq) & (specFreqs <= maxPlotFreq)
		specSpectrum = specSpectrum[mask,:]  # TRANSPOSE
		'''
		print('  2 Transpose specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
		print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
		'''

		# careful: I want to eventually scale y-axis to [0, self.maxPlotHz]
		# for now use [minPlotFreq, maxPlotFreq]
		x_min = startSec
		x_max = stopSec
		y_min = maxPlotFreq  # minPlotFreq
		y_max = minPlotFreq  # maxPlotFreq
		extent = [x_min , x_max, y_min , y_max]

		self.spectrogramAxes.clear()
		self.spectrogramAxes.imshow(specSpectrum, extent=extent, aspect='auto', cmap='viridis')
		self.spectrogramAxes.invert_yaxis()
		self.spectrogramAxes.set_xlabel('Time (s)')
		self.spectrogramAxes.set_ylabel('Freq (Hz)')

		#
		# matplotlib psd
		if self.psdWindowStr == 'Hanning':
			psdWindow = mpl.mlab.window_hanning
		elif self.psdWindowStr == 'Blackman':
			psdWindow = np.blackman(nfft)
		else:
			logger.warning(f'psdWindowStr not understood {self.psdWindowStr}')
		'''
		print('=== calling mpl psd()')
		print('  nfft:', nfft)
		print('  fs:', fs)
		print('  myDetrend:', myDetrend)
		print('  psdWindowStr:', self.psdWindowStr)
		print('  psdWindow:', psdWindow)
		'''
		# default scale_by_freq=True
		scale_by_freq = True
		Pxx, freqs = self.psdAxes.psd(yFiltered, marker='', linestyle='-', NFFT=nfft, Fs=fs,
									scale_by_freq=scale_by_freq, detrend=self.linearDetrend, window=psdWindow)

		#
		# replot matplotlib psd
		pxxLog10 = 10 * np.log10(Pxx)
		
		#pxxLog10 = Pxx # does this look better ???
		
		mask = (freqs >= minPlotFreq) & (freqs <= maxPlotFreq)
		pxxLog10 = pxxLog10[mask] # order matters
		freqsLog10 = freqs[mask] # order matters
		self.psdAxes.clear()
		self.psdAxes.plot(freqsLog10, pxxLog10, '-', linewidth=1)
		self.psdAxes.set_xlim([0, self.maxPlotHz])  # x-axes is frequency
		self.psdAxes.grid(True)
		#self.psdAxes.set_ylabel('10*log10(Pxx)')
		self.psdAxes.set_ylabel('PSD (dB)')
		self.psdAxes.set_xlabel('Freq (Hz)')

		#
		# get peak frequency from psd, finding max peak with width
		inflection = np.diff(np.sign(np.diff(pxxLog10)))
		peaks = (inflection < 0).nonzero()[0] + 1
		# excception ValueError
		try:
			peak0 = peaks[pxxLog10[peaks].argmax()]
			maxFreq0 = round(freqsLog10[peak0],3) # Gives 0.05
			maxPsd0 = round(pxxLog10[peak0],3) # Gives 0.05
		except(ValueError) as e:
			logger.error('BAD PEAK in pxx')
			maxFreq0 = []
			maxPsd0 =[]
		#print(f'1 Results: Peak Hz={round(maxFreq0,3)} Amplitude={round(maxPsd0,3)}')

		# add to plot
		self.psdAxes.plot(maxFreq0, maxPsd0, '.r')

		# the absolute maximum in the psd
		maxPsd = np.nanmax(pxxLog10)
		maxPnt = np.argmax(pxxLog10)
		maxFreq = freqsLog10[maxPnt]
		#self.resultsLabel.setText(f'Results: Peak Hz={round(maxFreq,3)} Amplitude={round(maxPsd,3)}')
		#print(f'2 Results: Peak Hz={round(maxFreq,3)} Amplitude={round(maxPsd,3)}')
 
		#
		# print results
		# (file, startSec, stopSec, max psd amp, max psd freq)
		pStart = round(startSec,3)
		pStop = round(stopSec,3)
		pMaxFreq = round(maxFreq,3)
		pMaxPsd = round(maxPsd,3)

		'''
		printStr = f'Type\tFile\tstartSec\tstopSec\tmaxFreqPsd\tmaxPsd'  #\tmaxFreqFft\tmaxFft'
		#self.appendResultsStr(printStr)
		print('=== FFT results are:')
		print(printStr)
		printStr = f'fftPlugin\t{self.ba.getFileName()}\t{pStart}\t{pStop}\t{pMaxFreq}\t{pMaxPsd}'
		self.appendResultsStr(printStr, maxFreq=pMaxFreq, maxPsd=pMaxPsd, 
								maxFreq0=maxFreq0, maxPsd0=maxPsd0,
								freqs=freqsLog10, psd=pxxLog10)  # last 2 are not used
		print(printStr)
		'''

		# 20220330, rms error of stim
		tmpSweepC = sweepC[leftPoint:rightPoint]
		rms = np.sqrt(np.mean(tmpSweepC**2))

		resultDict = {
			'file': self.ba.getFileName(),
			'sweep': self.ba.currentSweep,
			'startSec': self.getStartStop()[0],
			'stopSec': self.getStartStop()[1],
			'butterFilter': self.doButterFilter,
			'butterOrder': self.butterOrder,
			'lowFreqCutoff': self.butterCutoff[0],
			'highFreqCutoff': self.butterCutoff[1],
			# ultimate max
			'maxFreq': pMaxFreq,
			'maxPSD': pMaxPsd,
			# max in peak
			'maxFreq0': maxFreq0,
			'maxPSD0': maxPsd0,
			# TODO: max as stim freq
			'maxFreq_at_stim': '',
			'maxPSD_at_stim': '',
			'rmsStim': rms,  # rms of injected current
			#'freqs': freqs,
			#'psd': psd,
		}
		self.appendResults2(resultDict)

		# TODO: make a dict of results and append to pandas df

		#
		# plot np fft
		# see: https://www.gw-openscience.org/tutorial05/
		# blackman is supposed to correct for low freq signal ???
		'''
		doBlackman = False
		if doBlackman:
			window = np.blackman(yFiltered.size)
			windowed_yFiltered = yFiltered*window
		else:
			windowed_yFiltered = yFiltered

		ft = np.fft.rfft(windowed_yFiltered)
		# not sure to use diff b/w [1]-[0] or fs=10000 ???
		tmpFs = t[1]-t[0]
		fftFreqs = np.fft.rfftfreq(len(windowed_yFiltered), tmpFs) # Get frequency axis from the time axis
		fftMags = abs(ft) # We don't care about the phase information here

		# find max peak with width
		inflection = np.diff(np.sign(np.diff(fftMags)))
		peaks = (inflection < 0).nonzero()[0] + 1
		peak = peaks[fftMags[peaks].argmax()]
		signal_freq = round(fftFreqs[peak],3) # Gives 0.05
		signal_mag = round(fftMags[peak],3) # Gives 0.05
		#printStr = f'FFT\t{self.ba.getFileName()}\t{pStart}\t{pStop}\t{signal_freq}\t{signal_mag}'
		#print(printStr)
		printStr += f'\t{signal_freq}\t{signal_mag}'
		print(printStr)
		self.appendResultsStr(printStr)
		#print(f'  fft signal frequency is:{signal_freq} with mag:{signal_mag}')

		# strip down to min/ax x-plot of freq
		mask = (fftFreqs>=minPlotFreq) & (fftFreqs<=maxPlotFreq)
		fftMags = fftMags[mask] # order matters
		fftFreqs = fftFreqs[mask] # order matters

		fftMags = fftMags[2:-1]
		fftFreqs = fftFreqs[2:-1]

		self.fftAxes.clear()
		self.fftAxes.semilogy(fftFreqs, fftMags, '-', linewidth=1)
		self.fftAxes.set_xlim([minPlotFreq, maxPlotFreq])  # x-axes is frequency
		self.fftAxes.set_xlabel('Freq (Hz)')
		self.fftAxes.set_ylabel('FFT Mag')
		self.fftAxes.grid(True)
		self.fftAxes.semilogy(signal_freq, signal_mag, '.r')
		'''

		stop = time.time()
		#logger.info(f'Took {stop-start} seconds.')

		#
		#self.static_canvas.draw()

		figFileName = self.getFigFileName()
		self.fig.suptitle(figFileName)

	def appendResults2(self, resultDict):
		self._resultsDictList2.append(resultDict)

	def appendResultsStr(self, str, maxFreq='', maxPsd='', 
							maxFreq0='', maxPsd0='',
							freqs='', psd=''):
		self._resultStr += str + '\n'
		resultDict = {
			'file': self.ba.getFileName(),
			'sweep': self.ba.currentSweep,
			'startSec': self.getStartStop()[0],
			'stopSec': self.getStartStop()[1],
			'butterFilter': self.doButterFilter,
			'butterOrder': self.butterOrder,
			'lowFreqCutoff': self.butterCutoff[0],
			'highFreqCutoff': self.butterCutoff[1],
			# ultimate max
			'maxFreq': maxFreq,
			'maxPSD': maxPsd,
			# max in peak
			'maxFreq0': maxFreq0,
			'maxPSD0': maxPsd0,
			#'freqs': freqs,
			#'psd': psd,
		}
		self._resultsDictList.append(resultDict)

	def getResultsDictList(self):
		return self._resultsDictList

	def getResultStr(self):
		return self._resultStr

def run():
	# need to use colinAnalysis.bAnalysis2() to load stim file atf
	# see how far I get without it, all I need from it is
	#	- (sin, noise) params per sweep
	#	- if a sweep is blank_re
	# 	- try and scrip these things into a table and just use a regular bAnalysis

	#
	# if we were recorded with a stimulus file abf
	# needed to assign stimulusWaveformFromFile
	#tmpSweepC = ba.abf.sweepC

	from pprint import pprint

	masterDf = pd.DataFrame()
	saveFolder = '/Users/cudmore/Desktop/stoch-plots'

	folderPath = '/Users/cudmore/Library/CloudStorage/Box-Box/data/stoch-res/new20220104'

	from sanpy.bAnalysisStim import bAnalysisStim

	for file in sorted(os.listdir(folderPath)):
		if not file.endswith('.abf'):
			continue
		
		fullPath = os.path.join(folderPath, file)
		
		ba = bAnalysisStim(fullPath)  # bAnalysis that loads my stim abf (with header from stimgen)	
		#ba = sanpy.bAnalysis(fullPath)

		if ba.numSweeps < 2:
			print('!!! skipping file:', file)
			continue

		print(ba)

		dfStimFile = ba._stimFileDf
		pprint(dfStimFile)

		sweep = 0
		fftRecording = fftFromRecording(ba=ba, sweepNumber=sweep,
								startSec = 5, stopSec = 25,
								doPlot=False)

		fftRecording.medianFilterPnts = 5

		for sweep in range(ba.numSweeps):
			print(f'{sweep} {ba}')
			fftRecording.replot_fft2(sweepNumber=sweep)
			fftRecording.saveFig(folderPath=saveFolder)

		#resultsDictList = fftRecording.getResultsDictList()
		resultsDictList = fftRecording._resultsDictList2
		df = pd.DataFrame(resultsDictList)

		# add stim params to each sweep
		df['stimFreq'] = dfStimFile['freq(Hz)']
		df['noiseAmp'] = dfStimFile['noise amp']

		masterDf = pd.concat([masterDf, df], ignore_index=True)

	#
	pprint(masterDf)

	summaryFile = os.path.join(saveFolder, 'stoch-summary.csv')
	print('saving:', summaryFile)
	masterDf.to_csv(summaryFile)

	#plt.show()

if __name__ == '__main__':
	run()
	



# 20210129

import os, sys
import numpy as np

import tifffile

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class bAbfText:
    """
    mimic a pyabf file to load from a text file
    """
    def __init__(self, path=None, theDict=None):
        """
        path is either:
            path to .csv file with columns (time (seconds), vm)
            path to .tif file

        theDict:
            sweepX
            sweepY
        """

        self.path = path

        self.sweepX = None
        self.sweepY = None
        self.dataRate = None
        self.dataPointsPerMs = None
        self.sweepList = [0]

        self.tif = None
        self.tifNorm = None # not used
        self.tifHeader = None

        if theDict is not None:
            logger.info('bAbfText() from dict')
            self.sweepX = theDict['sweepX']
            self.sweepY = theDict['sweepY']
        elif path.endswith('.tif'):
            self.sweepX, self.sweepY, self.tif, self.tifNorm = \
                                self._abfFromLineScanTif(path)
            #self.tifHeader = self._loadLineScanHeader(path)

            #logger.info(f'tif params: {np.min(self.tif)} {np.max(self.tif)} {self.tif.shape} {self.tif.dtype}')
        else:
            logger.info('bAbfText() from path: {path}')
            if not os.path.isfile(path):
                logger.error('Did not find file: {path}')
                return

            tmpNumPy = np.loadtxt(path, skiprows=1, delimiter=',')
            self.sweepX = tmpNumPy[:,0]
            self.sweepY = tmpNumPy[:,1]

        #
        # linescan is like
        # 'secondsPerLine': 0.0042494285714285715,
        # 1 / secondsPerLine = 235.32
        # 20220926, if this is '1' second, detection will fail
        #  because self.dataPointsPerMs goes to 0
        secondsPerSample = self.sweepX[1] - self.sweepX[0]
        samplesPerSecond = 1 / secondsPerSample
        self.dataRate = samplesPerSecond #235 #10000 # samples per second

        # todo: calculat from dataRate
        secondsPerSample = 1 / samplesPerSecond
        msPerSample = secondsPerSample * 1000
        samplesPerMs = 1 / msPerSample
        self.dataPointsPerMs = samplesPerMs #4.255 #1/235*1000,10 # 10 kHz

        self.sweepUnitsX = 'todo: fix'
        self.sweepUnitsY = 'todo: fix'

        #self.myRectRoi = None
        self.myRectRoiBackground = None
        #self.minLineScan = None

        # done in _abfFromLineScanTif
        # defaultRect = self.defaultTifRoi_background()  # for kymograph
        # self.updateTifRoi_background(defaultRect)

        # defaultRect = self.defaultTifRoi()  # for kymograph
        # self.updateTifRoi(defaultRect)

        #
        # print results
        '''
        if self.tifHeader is not None:
            for k,v in self.tifHeader.items():
                print('  ', k, ':', v)

        print('  ', 'bAbfText secondsPerSample:', secondsPerSample, 'seconds/sample')
        print('  ', 'bAbfText self.dataRate:', self.dataRate, 'samples/second')
        print('  ', 'bAbfText self.dataPointsPerMs:', self.dataPointsPerMs)
        '''

    def defaultTifRoi_background(self):
        """
        Default ROI for kymograph background

        - height/width about 5 pixels
        - position near the end of the line
        """
        widthTif = self.tif.shape[1]
        heightTif = self.tif.shape[0]
        insetPixels = 10  # from end of line (top)
        roiWidthHeight = 5
        #tifHeightPercent = self.tif.shape[0] * 0.15

        left = int(widthTif/2)
        top = heightTif - insetPixels
        right = left + roiWidthHeight - 1
        bottom = top - roiWidthHeight + 1

        return [left, top, right, bottom]

    def resetRoi(self):
        """reset the kymograph roi to default
        """
        defaultRect = self.defaultTifRoi()
        self.updateTifRoi(defaultRect)
        return defaultRect

    def defaultTifRoi(self):
        """
        A default rectangular ROI for main kymograph analysis.

        TODO: Remove hard coding of 0.15 percent of height
        
        Return:
            list: [left, top, right, bottom]
        """
        xRoiPos = 0  # startSeconds
        yRoiPos = 0  # pixels
        widthRoi = self.tif.shape[1]
        heightRoi = self.tif.shape[0]
        tifHeightPercent = self.tif.shape[0] * 0.15
        yRoiPos += tifHeightPercent
        heightRoi -= 2 * tifHeightPercent

        left = xRoiPos
        top = yRoiPos + heightRoi
        right = xRoiPos + widthRoi
        bottom = yRoiPos

        left = int(left)
        top = int(top)
        right = int(right)
        bottom = int(bottom)

        return [left, top, right, bottom]

    def getTifRoiBackground(self):
        return self.myRectRoiBackground

    def getTifRoi(self):
        return self.myRectRoi

    def updateTifRoi_background(self, theRect=None):
        if theRect is None:
            theRect = self.defaultTifRoi_background()
        if theRect is not None:
            left = theRect[0]
            top = theRect[1]
            right = theRect[2]
            bottom = theRect[3]

        self.myRectRoiBackground = theRect

        tif = self.tif
        theSum = np.sum(tif[bottom:top, left:right])

        return theSum

    def updateTifRoi(self, theRect=None):
        """
        Recalculate (sweepX, sweepY) based on user specified rect ROI

        Args:
            theRect (list): left, top, right, bottom

        Notes:
            This is doing 'background subtract' on sum of line - sum in background roi
                I don't think this is correct???
                'sum of line' has num pixels in scan
                'sum background roi' has a different number of pixels !!!
        """

        logger.info(theRect)

        self.myRectRoi = theRect

        firstFrameSeconds = 0

        tif = self.tif

        left = 0  # startSec
        top = tif.shape[0]
        right = tif.shape[1]  # stopSec
        bottom = 0
        if theRect is not None:
            #left = theRect[0]
            top = theRect[1]
            #right = theRect[2]
            bottom = theRect[3]

        # number of pixels in subset of line scan
        numPixels = top - bottom + 1

        #theBackground = self.updateTifRoi_background()

        # logger.warning(f'I turned off background subtract .. placement of background is sometimes bad')

        _secondsPerLine = self.tifHeader['secondsPerLine']
        if _secondsPerLine is None:
            _secondsPerLine = 0.001  # so we get 1 pnt per ms
            logger.warning(f'No header, secondsPerLine is None. Using fake _secondsPerLine:{_secondsPerLine}')

        # retains shape of entire tif
        yLineScanSum = [np.nan] * tif.shape[1]
        xLineScanSum = [np.nan] * tif.shape[1]
        for i in range(tif.shape[1]):
            # sum each column and background subtract

            # always assign time
            xLineScanSum[i] = firstFrameSeconds + (_secondsPerLine * i)
            # if self.tifHeader['secondsPerLine'] is None:
            #     # no units on kymograph
            #     _fakeSecondsPerLine = 0.001
            #     #logger.error(f'No header, secondsPerLine is None. Using _fakeSecondsPerLine:{_fakeSecondsPerLine}')
            #     xLineScanSum[i] = firstFrameSeconds + (_fakeSecondsPerLine * i)
            # else:
            #     xLineScanSum[i] = firstFrameSeconds + (self.tifHeader['secondsPerLine'] * i)

            if i<left or i>right:
                continue

            # each line scan
            theSum = np.sum(tif[bottom:top, i])

            # background subtract
            #print(f'   === theSum:{theSum} {type(theSum)} theBackground:{theBackground} {type(theBackground)}')
            #if not isinstance(theSum, np.uint64) or not isinstance(theBackground, np.uint64):
            #    print(f'   === theSum:{theSum} {type(theSum)} theBackground:{theBackground} {type(theBackground)}')

            #logger.warning(f'I turned off background subtract .. placement of background is sometimes bad')
            #theSum -= theBackground

            #sys.exit(1)

            # f / f_0
            theSum /= numPixels  # tif.shape[0] # norm to number of pixels in line
            yLineScanSum[i] = theSum

        xLineScanSum = np.asarray(xLineScanSum)
        yLineScanSum = np.asarray(yLineScanSum)

        # derive simplified f/f0 by dividing by min
        self.minLineScan = np.nanmin(yLineScanSum)
        
        # 20221003, do not do this, ysweep will be sum/pnts
        # yLineScanSum /= self.minLineScan

        logger.info('updating sweepX and sweepY with new rect ROI')
        self.sweepX = xLineScanSum
        self.sweepY = yLineScanSum

    def setSweep(self, sweep):
        pass

    def _abfFromLineScanTif(self, path, theRect=None):
        """

        Returns:
            xLineScanSum (np.ndarray) e.g. sweepX
            yLineScanSum (np.ndarray) e.g. sweepY
        """
        if not os.path.isfile(path):
            logger.error(f'Did not find file: {path}')
            return None, None, None, None

        tif = tifffile.imread(path)
        logger.info(f'Loaded tif with shape: {tif.shape}')
        
        if len(tif.shape) == 3:
            tif = tif[:,:,1] # assuming image channel is 1
        
        # we want our tiffs to be short and wide
        # shape[0] is space, shape[1] is time
        if tif.shape[0] < tif.shape[1]:
            # correct shape
            pass
        else:
            tif = np.rot90(tif) # rotates 90 degrees counter-clockwise

        f0 = tif.mean()
        tifNorm = (tif - f0) / f0

        self.tif = tif

        # assuming this was exported using Olympus software
        # which gives us a .txt file
        self.tifHeader = self._loadLineScanHeader(path)
        #self.tifHeader['shape'] = tif.shape
        #self.tifHeader['secondsPerLine'] = \
        #                self.tifHeader['totalSeconds'] / self.tifHeader['shape'][1]
        #tifHeader['abfPath'] = abfFile

        # updateTifRoi sets sweepX and seepY
        defaultRect = self.defaultTifRoi()
        self.updateTifRoi(defaultRect)

        firstFrameSeconds = 0

        #
        # sum the inensities of each image line scan
        #print('tif.shape:', tif.shape) # like (146, 10000)
        '''
        yLineScanSum = [np.nan] * tif.shape[1]
        xLineScanSum = [np.nan] * tif.shape[1]
        for i in range(tif.shape[1]):
            theSum = np.sum(tif[:,i])
            theSum /= tif.shape[0] # norm to number of pixels in line
            yLineScanSum[i] = theSum
            xLineScanSum[i] = firstFrameSeconds + (self.tifHeader['secondsPerLine'] * i)

        xLineScanSum = np.asarray(xLineScanSum)
        yLineScanSum = np.asarray(yLineScanSum)
        '''

        # normalize to 0..1
        # 20220114, was this
        # yLineScanSum = self._NormalizeData(yLineScanSum)

        #print('bAbfText xLineScanSum:', xLineScanSum.shape)
        #print('bAbfText yLineScanSum:', yLineScanSum.shape)
        
        return self.sweepX, self.sweepY, tif, tifNorm

    def _loadLineScanHeader(self, path):
        """Load scale from .txt file saved by olympus export.

        Args:
            path: full path to tif

        returns dict:
            numPixels:
            umLength:
            umPerPixel:
            totalSeconds:
        """
        def _defaultHeader():
            defaultHeader = {
                'tif': path,
                'numLines': self.tif.shape[1],
                'totalSeconds': None,
                'shape': self.tif.shape,
                'secondsPerLine': 0.001,
                #'linesPerSecond': None,  # 1/secondsPerLine

                'numPixels': self.tif.shape[0],
                'umLength': None,
                'umPerPixel': 1,
            }
            return defaultHeader
        
        # "X Dimension"    "138, 0.0 - 57.176 [um], 0.414 [um/pixel]"
        # "T Dimension"    "1, 0.000 - 35.496 [s], Interval FreeRun"
        # "Image Size(Unit Converted)"    "57.176 [um] * 35500.000 [ms]"
        txtFile = os.path.splitext(path)[0] + '.txt'

        theRet = _defaultHeader()
        #theRet = {'tif': path}
        #theRet['numLines'] = self.tif.shape[1]

        if not os.path.isfile(txtFile):
            logger.warning(f'Did not find file: {txtFile}')
            logger.warning(f'  The Kymograph will have no scale.')
            return theRet

        # to open export .txt file for Olympus FV3000
        with open(txtFile, 'r') as fp:
            lines = fp.readlines()
            for line in lines:
                line = line.strip()
                if line.startswith('"X Dimension"'):
                    line = line.replace('"', "")
                    line = line.replace(',', "")
                    #print('loadLineScanHeader:', line)
                    # 2 number of pixels in line
                    # 5 um length of line
                    # 7 um/pixel
                    splitLine = line.split()
                    for idx, split in enumerate(splitLine):
                        #print('  ', idx, split)
                        if idx == 2:
                            numPixels = int(split)
                            theRet['numPixels'] = numPixels
                        elif idx == 5:
                            umLength = float(split)
                            theRet['umLength'] = umLength
                        elif idx == 7:
                            umPerPixel = float(split)
                            theRet['umPerPixel'] = umPerPixel

                elif line.startswith('"T Dimension"'):
                    line = line.replace('"', "")
                    line = line.replace(',', "")
                    #print('loadLineScanHeader:', line)
                    # 5 total duration of image acquisition (seconds)
                    splitLine = line.split()
                    for idx, split in enumerate(splitLine):
                        #print('  ', idx, split)
                        if idx == 5:
                            totalSeconds = float(split)
                            theRet['totalSeconds'] = totalSeconds

                            #theRet['secondsPerLine'] =

                #elif line.startswith('"Image Size(Unit Converted)"'):
                #    print('loadLineScanHeader:', line)

        #theRet['shape'] = self.tif.shape
        theRet['secondsPerLine'] = theRet['totalSeconds'] / theRet['shape'][1]
        #theRet['linesPerSecond'] = 1 / theRet['secondsPerLine']
        #
        return theRet

    def _NormalizeData(self, data):
        """
        normalize to [0..1]
        """
        return (data - np.min(data)) / (np.max(data) - np.min(data))

def test_load_tif():
    path = '/Users/cudmore/data/dual-lcr/20210115/data/20210115__0001.tif'
    abf = bAbfText(path=path)

    import matplotlib.pyplot as plt
    plt.plot(abf.sweepX, abf.sweepY)
    plt.show()

if __name__ == '__main__':

    test_load_tif()
